"""Script to simulate an Ecological Momentary Assessment (EMA) study.
Collects all simulated EMA records in xlsx files,
just as if the data had been gathered in a real experiment,
and runs the analysis.

A simulated experiment may include
one or more Groups of subjects randomly drawn from one or more Population(s),
with each subject performing several simulated EMA recordings.
Each EMA record includes
(1) a nominal SCENARIO, specified in one or more scenario dimensions,
(2) ordinal ratings of zero, one or more perceptual ATTRIBUTE(s).


*** Usage:
Copy and edit this script and run it for any desired simulation,
or run it unchanged, as a demo of the EmaCalc package.

*** Version history:

* Version 0.7.1:
2022-01-19, allow set_random_seed for exactly reproducible results

* Version 0.7: minor update to include new calculation and display functions

* Version 0.6:
2021-12-08, tested methods for restrict_attribute, restrict_threshold

* Version 0.5:
2021-11-16, first functional version
2021-11-21, cleaned and tested beta version
"""
import numpy as np
from pathlib import Path
import logging
import pickle

from EmaCalc.ema_simulation import EmaSimPopulation, EmaSimExperiment
from EmaCalc.ema_simulation import SubjectThurstone, SubjectBradley

from EmaCalc.ema_data import EmaFrame, EmaDataSet
from EmaCalc.ema_model import EmaModel
from EmaCalc.ema_display import EmaDisplaySet
from EmaCalc import ema_logging, __version__

# ------------ Set simulator random seed for reproducible results
# from EmaCalc.ema_simulation import set_sim_seed
# set_sim_seed(12345)  # ONLY FOR TEST OR DEMO

# -------- __main__ check to prevent multiprocessor sub-tasks to re-run this script
if __name__ == '__main__':
    # ------------------------ Set up working directory and result logging:
    work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
    data_path = work_path / 'data'  # or whatever
    result_path = work_path / 'result'  # or whatever

    model_file = 'test_ema_model.pkl'  # saved model file
    experiment_file = 'test_ema_experiment.pkl'  # saved EmaSimExperiment

    ema_logging.setup(save_path=result_path,
                      log_file='run_sim_log.txt')  # to save the log file

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    logger.info(f'*** Running EmaCalc version {__version__} ***')

    # --------------- Define Experimental Framework: Scenarios, Attributes, and Grades

    emf = EmaFrame(scenarios={'Stage': ('',),   # Test Stage(s)
                              'HA': ('A', 'B'),  # two Hearing-aid programs
                              'CoSS': [f'{i}.' for i in range(1, 8)],   # Seven CoSS categories
                              },  # nominal categories
                   stage_key='Stage',
                   attribute_grades={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},  # ordinal rating categories
                   )
    # NOTE: First scenario dimension is always Stage, even if only ONE
    # User may set arbitrary stage_key label

    # --------------- Define Population(s) to be simulated:

    # Not necessarily normalized probability-mass arrays
    pop0_scenario_prob = np.ones((2, 7))  # with HA = A, B

    # pop0_scenario_prob.shape == emf.scenario_shape
    # OR == emf.scenario_shape[1:] if emf.scenario_shape[0] == 1, i.e., only one Stage

    # --------- Population mean for Attribute 'Speech', in subject_class scale units

    pop0_speech_mean = np.array([[0., 0., 0., 0., 0., 0., 0.],     # with 'HA' = 'A'
                                 [3., 3., 3., 0., -3., -3., -3]])  # with 'HA' = 'B'
    pop0_speech_mean[1] += 0.5  # add fixed difference between HA (B - A)
    pop0_comfort_mean = pop0_speech_mean[[1, 0], :]  # opposite pattern to 'Speech'

    pop_0 = EmaSimPopulation(emf,
                             scenario_prob=pop0_scenario_prob,
                             attribute_mean={'Speech': pop0_speech_mean,
                                             'Comfort': pop0_comfort_mean},
                             log_scenario_std=0.,  # inter-individual standard deviation of log prob
                             attribute_std=0.5,  # inter-individual standard deviation of attribute locations
                             response_width_mean=1.5,  # response interval width in subject_class units
                             log_response_width_std=0.3,  # inter-individual random threshold variations
                             # subject_class=SubjectBradley,  # default
                             id='Pop0'
                             )
    # Optionally, define other populations here,
    pop1_scenario_prob = pop0_scenario_prob
    pop1_speech_mean = np.array([[0., 0., 0., 0., 0., 0., 0.],     # with 'HA' = 'A'
                                 [0., 0., 0., 0., 0., 0., 0.]])  # with 'HA' = 'B'
    pop1_speech_mean[0] -= 1.  # add fixed difference between HA (B - A)
    pop1_speech_mean[1] += 1.  # add fixed difference between HA (B - A)
    pop1_comfort_mean = pop1_speech_mean[[1, 0], :]  # opposite pattern to 'Speech'
    # pop1_comfort_mean = pop0_comfort_mean  # SAME as pop0
    pop_1 = EmaSimPopulation(emf,
                             scenario_prob=pop1_scenario_prob,
                             attribute_mean={'Speech': pop1_speech_mean,
                                             'Comfort': pop1_comfort_mean},
                             log_scenario_std=0.,  # inter-individual standard deviation of log prob
                             attribute_std=0.5,  # inter-individual standard deviation of attribute locations
                             response_width_mean=1.5,  # response interval width in subject_class units
                             log_response_width_std=0.3,  # inter-individual random threshold variations
                             # subject_class=SubjectBradley,  # default
                             id='Pop1'
                             ) # with SAME EmaFrame, but different scenario_prob and/or mean Attribute values

    # ------------------ Generate Experiment with one Group of Subjects from each Population:
    ema_exp = EmaSimExperiment(emf,
                               groups={(('Age', 'old'),): pop_0.gen_group(n_subjects=20),
                                       (('Age', 'young'),): pop_1.gen_group(n_subjects=15)
                                       })
    # NOTE: groups keys MUST be a tuple of ONE OR MORE pairs (g_factor, g_cat)
    #
    logger.info(f'ema_exp= {ema_exp}')
    # ema_exp defines true parameter values for Population(s) and all Subjects in each Group

    # -------- (Optional) Save EmaSimExperiment instance with true subject parameters
    with (work_path / experiment_file).open('wb') as f:
        pickle.dump(ema_exp, f)

    # ------------------------- Generate EMA data for all subjects in all Groups:

    ds = ema_exp.gen_dataset(min_ema=30,    # min n of EMA records per subject
                             max_ema=70)    # max n of EMA records per subject
    # = a complete EmaDataSet instance with simulated EMA records for all subjects

    # Optionally, save all data in a set of data files, one for each subject
    file_format = 'xlsx'
    ds.save(data_path, allow_over_write=True, fmt=file_format,
            subject='sheet',    # subject ID stored in xlsx sheet title
            ema_vars={'Stage': 'A',  # column address
                      'HA': 'B',
                      'CoSS': 'C',
                      'Speech': 'F',
                      'Comfort': 'G'}
            )
    # -> one xlsx file for each subject
    logging.info('Simulated EMA data saved in ' + str(data_path) + f' as {file_format} files')

    # ------------------------------- Learn Analysis Model from simulated data set:

    # -------- Test re-loading simulated data from saved xlsx files:
    logging.info('Simulated EMA data loaded from ' + str(data_path) + f' as {file_format} files')
    ds = EmaDataSet.load(emf, data_path,
                         # grouping=None,  # default: include all subjects as ONE unnamed group
                         grouping={'Age': ('old',),  # analyze only Age=old
                                   },
                         # grouping={'Age': ('young','old')},  # analyze both Age groups separately
                         subject='sheet',
                         ema_vars={'HA': 'B',
                                   'CoSS': 'C',
                                   'Speech': 'F',
                                   'Comfort': 'G'}
                         )

    logger.info(f'Using data ds=\n{ds}')

    # Model ordinal-regression effects of Scenarios on each Attribute:
    # regression_effects = ['HA',     # main linear regression effect only
    #                       'CoSS',   # main linear regression effect only
    #                       # 'Stage',  # if there are several Stage categories
    #                       ]
    regression_effects = [('HA', 'CoSS')  # joint effects, main AND interaction
                          # 'Stage',  # if there are several Stage categories
                          ]
    # NOTE: regression effects may include any combination of scenario dimensions, but
    # including ALL interactions -> many model parameters,
    # possibly -> less precise estimation for each parameter.

    # In this example: ['HA', 'CoSS'] -> 2 + (7 - 1) = 8 regression-effect parameters
    #                ['CoSS', 'HA'] -> 7 + (2 - 1) = 8 regression-effect parameters
    #                [('HA', 'CoSS')] -> 2 * 7 = 14 regression-effect parameters

    emm = EmaModel.initialize(ds,
                              effects=regression_effects,
                              max_n_comp=10,
                              restrict_attribute=False,  # default
                              restrict_threshold=True,   # default
                              # seed=12345  # ONLY if reproducible results are required
                              )
    # max_n_comp = max number of mixture components in population model
    # restrict_attribute=True -> force sensory-variable locations centered at zero
    # restrict_threshold=True -> force response thresholds centered at zero
    # for each respondent and each sample of each attribute

    ll = emm.learn(max_hours=2., max_minutes=0.)
    logger.info('ll= ' + np.array2string(np.array(ll), precision=5))
    emm.prune()  # keep only active mixture components

    # -------- Save learned EmaModel instance (optional):
    with (work_path / model_file).open('wb') as f:
        pickle.dump(emm, f)

    # ------------------------------- generate all result displays:

    emd = EmaDisplaySet.show(emm,
                             scenarios=['CoSS',  # CoSS profile, aggregated across HA
                                        ('CoSS', 'HA'),  # CoSS profiles, conditional on HA
                                        ('HA', 'CoSS'),  # HA profiles, conditional on CoSS
                                        ],
                             attributes=[('Speech', 'CoSS'),     # Speech, main effect of CoSS
                                         ('Speech', 'HA'),       # Speech, main effect of HA
                                         ('Speech', ('CoSS', 'HA')),    # joint effect of both
                                         ('Comfort', ('CoSS', 'HA'))],  # joint effect of both
                             population_individual=True,  # random individual in the population
                             population_mean=True,  # population mean
                             subjects=False,  # individual results: True -> MANY plots and tables
                             nap=[('Speech', 'HA'),  # NAP(HA) grades aggregated across CoSS
                                  ('Speech', ('HA', 'CoSS')),
                                  ('Comfort', 'HA')],
                             nap_confidence=0.95,  # level of symmetric confidence interval
                             grade_thresholds=True,  # response thresholds in attribute plots
                             percentiles=[2.5, 25, 50, 75, 97.5],  # in profile plots and tables
                             credibility_limit=0.7,  # minimum credibility in difference tables
                             figure_format='pdf',  # or 'eps', 'jpg'
                             table_format='latex',  # or 'tab' for tab-limited text style
                             )
    # NOTE:
    # joint (=interaction) effects are correct only if included in EmaModel effects
    # nap results calculated only for Scenario dimension with exactly TWO categories.

    # ------------------------------- save all displays in directory tree
    emd.save(result_path)

    logging.info(f'All result displays saved in {result_path}')

    logging.shutdown()



"""This module defines classes and functions to display analysis results
given an EmaModel instance,
learned from a set of data from an EMA study.

Results are shown as figures and tables.
Figures can be saved in various formats, as specified in arguments to display method.
Tables are saved in LaTeX tabular format OR in tab-delimited text files.
Thus, both figures and tables can be easily imported into a LaTeX or word-processing document.

*** Main Class:

EmaDisplaySet = a structured container for all display results

Each display element can be accessed and modified by the user, before saving.

The display set can include data for three types of predictive distributions:
*: a random individual in the (Sub-)Population from which a group of respondents were recruited,
    (most relevant for an experiment aiming to predict the success of a new product,
    among individual potential customers in a population)
*: the mean (=median) in the (Sub-)Population from which a group of test subjects were recruited
    (with interpretation most similar to a conventional significance test)
*: each individual respondent in a Group of test subjects, with observed EMA data


*** Usage Example:
    See main scripts run_ema and run_sim

Figures and tables are automatically assigned descriptive names,
and saved in sub-directories with names constructed from
string labels of Groups and Attributes and Scenarios,
as defined in the ema_data.EmaFrame object of the input ema_model.EmaModel instance.

If there is more than one Group,
    one sub-directory is created for each group,
    and one sub-sub-directory for each requested population / subject predictive result,
    and one sub-sub-sub-directory for each attribute, with results within the group.

Thus, after saving, the display files are stored in a directory structure as, e.g.,
result_path / group / 'population_individual' / attributes / ....
result_path / group / 'population_individual' / scenarios / ....
result_path / group / 'subjects' / subject_id / attributes / ....
result_path / group / 'subjects' / subject_id / scenarios / ....
result_path / group / 'subjects_NAP' / ....  (if user requested NAP results)
result_path / 'group_effects' / 'population_mean' / attributes / ...  (if more than one group)

*** Version History:
* Version 0.8:
2022-02-15, minor cleanup of scenario profile tables

* Version 0.7:
2021-12-19, include table(s) of individual subject NAP results
2021-12-17, display aggregated Attribute effects weighted by Scenario probabilities

* Version 0.6:
2021-12-08, minor checks against numerical overflow

* Version 0.5.1
2021-11-27, allow NO Attributes in model, check display requests, minor cleanup

* Version 0.5
2021-11-05, copied from PairedCompCalc, modified for EmaCalc
2021-11-09, first functional version
"""
# ***** clarify fig_comments and table_comments??
# ***** methods display -> create ?? ****
# ***** local superclass for pretty-printed repr() ?
# ********* allow several Attributes in single plot ?

import numpy as np
from pathlib import Path
import logging
from itertools import product
import string

from . import ema_display_format as fmt

from samppy.credibility import cred_diff, cred_group_diff

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST

# N_SAMPLES = 1000

# ---------------------------- Default display parameters
FMT = {'scenarios': [],     # list of scenario dimensions to display
       'attributes': [],    # list of (attribute, scenario_effect) to display
       'percentiles': [5., 50., 95.],  # allowing 1, 2, 3, or more percentiles
       'credibility_limit': 0.7,  # min probability of jointly credible differences
       'sc_probability': 'Probability',  # label in figs and tables
       'population_individual': True,  # show result for random individual in population
       'population_mean': True,  # show result for population mean
       'subjects': False,   # show results for each respondent
       'nap': [],   # sequence of (Attribute, Scenarios) NAP results to display
       'nap_confidence': 0.95,  # range for symmetric NAP confidence intervals
       # ******* OR define nap_confidence = (2.5, 97.5) OR use 'percentiles' ?
       'scale_unit': '',  # scale unit for attribute plot axis
       }
# = dict with format parameters that may be changed by user
# Other format parameters are handled in module ema_display_format


def set_format_param(**kwargs):
    """Set / modify format parameters for this module, and module ema_display_format
    Called before any displays are generated.
    :param kwargs: dict with any formatting variables
    :return: None
    """
    other_fmt = dict()
    for (k, v) in kwargs.items():
        k = k.lower()
        if k in FMT:
            FMT[k] = v
        else:
            other_fmt[k] = v
    FMT['percentiles'].sort()  # ensure increasing values
    fmt.set_format_param(**other_fmt)  # all remaining user-defined parameters


# -------------------------------- Main Module Function Interface
def display(ema_model, **kwargs):  # *** not needed ? ***
    """Main function to display default set of EMA analysis results.
    :param: ema_model: a single EmaModel object
    :param: kwargs: (optional) any user-defined display format parameters
    :return: single QualityDisplaySet instance with display results
    """
    return EmaDisplaySet.show(ema_model, **kwargs)


# ------------------------------------------------ Elementary Display Class:
class Profile:
    """Container for ONE selected profile display
    for ONE tuple of one or more scenario dimensions,
    OR for ONE (attribute, scenario_effect)
    """
    def __init__(self, plot=None, tab=None, diff=None):
        """
        :param plot: (optional) FigureRef with profile plot
        :param tab: (optional) TableRef with plotted results tabulated.
        :param diff: (optional) TableRef with credible differences between profile elements.
        """
        self.plot = plot
        self.tab = tab
        self.diff = diff

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__) +
                '\n\t)')

    def save(self, dir_path):
        """
        :param dir_path: path to directory for saving files
        :return: None
        """
        if self.plot is not None:
            self.plot.save(dir_path)
        if self.tab is not None:
            self.tab.save(dir_path)
        if self.diff is not None:
            self.diff.save(dir_path)


# ---------------------------------------------------------- Top Display Class:
class EmaDisplaySet:
    """Root container for all displays
    of selected predictive scenario and attribute results
    from one ema_model.EmaModel instance learned from an ema_data.EmaDataSet instance.

    All display elements can be saved as files in a selected directory three.
    The complete instance can also be serialized and dumped to a single pickle file,
    then re-loaded, edited, and saved, if user needs to modify some display object.
    """
    def __init__(self, groups, group_effects=None):
        """
        :param groups: dict with (group: GroupDisplaySet) elements
            sub-divided among Attributes
        :param group_effects: (optional) single GroupEffectSet instance,
            showing jointly credible differences between groups,
            if there is more than one group
        """
        # *** store group_effects as dict with (a_label: GroupEffectSet object) ???
        self.groups = groups
        self.group_effects = group_effects

    def __repr__(self):  # *** general superclass for repr?
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    def save(self, dir_top):
        """Save all displays in a directory tree
        :param dir_top: Path or string with top directory for all displays
        :return: None
        """
        dir_top = Path(dir_top)  # just in case
        for (g, g_display) in self.groups.items():
            g = _dir_name(g, sep='/')
            if len(g) == 0 or all(s in string.whitespace for s in g):
                g_display.save(dir_top)
            else:
                g_display.save(dir_top / g)
        if self.group_effects is not None:
            self.group_effects.save(dir_top / 'group_effects')

    @classmethod
    def show(cls, emm,
             scenarios=None,
             attributes=None,
             nap=None,
             **kwargs):
        """Create displays for all results from an EMA study,
        and store all display elements in a single structured object.
        :param emm: single ema_model.EmaModel instance, with
            emm.groups[group] = ema_model.EmaGroupModel instance,
            emm.groups[group][subject_id] = ema_model.EmaSubjectModel instance
        :param scenarios: (optional) list with selected scenario dimensions to be displayed.
            scenarios[i] = a selected key in emm.emf.scenarios, or a tuple of such keys.
        :param attributes: (optional) list with selected attribute displays to be displayed.
            attributes[i] = a tuple (attribute_key, sc_effect), where
                attribute_key is one key in emm.emf.attribute_grades,
                sc_effect is a scenario key in emm.emf.scenarios, or a tuple of such keys.
                A single key will yield the main effect of the named scenario dimension.
                An effect tuple will also show interaction effects between dimensions,
                IFF the model has been trained to estimate such interaction effects.
        :param nap: (optional) list of selected (attribute, scenario) for desired NAP results.
        :param: kwargs: (optional) dict with any other display formatting parameters
            either for ema_display.FMT or ema_display_format.FMT
        :return: a cls instance filled with display objects
        """
        # get default scale_unit from emm, if not in kwargs
        if 'scale_unit' not in kwargs:
            kwargs['scale_unit'] = emm.base.rv_class.unit_label
        if scenarios is None:
            scenarios = [*emm.base.emf.scenarios.keys()]
            # default showing only main effects for each scenario dimension
        scenarios = [(sc,) if isinstance(sc, str) else sc
                     for sc in scenarios]
        # *** check that requested scenarios exist in model:
        missing_sc = [sc_i for sc_i in scenarios
                      if any(sc_ij not in emm.base.emf.scenarios.keys()
                             for sc_ij in sc_i)]
        for sc in missing_sc:
            logger.warning(f'Scenario profile {sc} unknown in the learned model.')
            scenarios.remove(sc)
        if attributes is None:
            attributes = []
        attributes = [(a, sc) if type(sc) is tuple else (a, (sc,))
                      for (a, sc) in attributes]
        # *** check that requested attribute effects exist in model
        missing_attr = [a_sc for a_sc in attributes
                        if a_sc[0] not in emm.base.emf.attribute_grades.keys()]
        for a_sc in missing_attr:
            logger.warning(f'Attribute effect {a_sc} unknown in the learned model.')
            attributes.remove(a_sc)
        if nap is None:
            nap = []
        else:
            nap = [(a, sc) if type(sc) is tuple else (a, (sc,))
                   for (a, sc) in nap]
        set_format_param(scenarios=scenarios,
                         attributes=attributes,
                         nap=nap,
                         **kwargs)
        # display separate results for each group
        groups = {g: GroupDisplaySet.display(emm_g)
                  for (g, emm_g) in emm.groups.items()}

        if len(groups) > 1:  # prepare for group differences
            group_effects = GroupEffectSet.display(emm)
        else:
            group_effects = None
        logger.info(fig_comments())
        logger.info(table_comments())
        return cls(groups, group_effects)


class GroupDisplaySet:
    """Container for all quality displays related to ONE experimental group:
    Predictive results for the population from which the subjects were recruited,
    and descriptive data for each individual subject, if requested by user.
    """
    def __init__(self,
                 population_mean=None,
                 population_individual=None,
                 subjects=None,
                 nap=None):
        """
        :param population_mean: dict with (a_label: AttributeDisplay object), for population mean,
        :param population_individual: dict with (a_label: AttributeDisplay object), for random individual,
        :param subjects: dict with (s_id: EmaDisplay instance) for participants in one group
            with scenario and attribute results.
        :param nap: dict with (nap_key: nap_table) elements,
            each with results for all subjects in this group
        """
        self.population_mean = population_mean
        self.population_individual = population_individual
        self.subjects = subjects
        self.nap = nap

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    def save(self, path):
        """Save all stored display objects in their corresponding sub-trees
        """
        if self.population_mean is not None:
            self.population_mean.save(path / 'population_mean')
        if self.population_individual is not None:
            self.population_individual.save(path / 'population_individual')
        if self.subjects is not None:
            for (s, s_disp) in self.subjects.items():
                s_disp.save(path / 'subjects' / str(s))
        if self.nap is not None:
            for (nap_key, nap_table) in self.nap.items():
                nap_table.save(path / 'subjects_NAP')

    @classmethod
    def display(cls, emm_g):
        """Generate all displays for ONE group
        :param emm_g: dict with elements (a_label: pc_result.PairedCompGroupModel)
        :return: cls instance with all displays for this group
        """
        pop_ind = None
        pop_mean = None
        subjects = None  # ***** empty dict in case no subjects shown
        # attr_corr = None
        if FMT['population_individual']:
            pop_ind = EmaDisplay.display(emm_g.predictive_population_ind())
        if FMT['population_mean']:
            pop_mean = EmaDisplay.display(emm_g.predictive_population_mean())
        if FMT['subjects']:
            # logger.debug('Displaying subjects:')
            subjects = {s: EmaDisplay.display(s_model)
                        for (s, s_model) in emm_g.subjects.items()}
        sc = emm_g.base.emf.scenarios
        nap = dict()
        for nap_sel in FMT['nap']:
            if len(sc[nap_sel[1][0]]) == 2:  # required for NAP calculation
                (attr, a_sc) = nap_sel
                nap[nap_sel] = fmt.nap_table({s: s_model.nap_diff(attr, a_sc,
                                                                  p=FMT['nap_confidence'])
                                              for (s, s_model) in emm_g.subjects.items()},
                                             attribute=attr,
                                             sc_diff=a_sc[0],
                                             case_labels={sc_dim: sc[sc_dim]
                                                          for sc_dim in a_sc[1:]},
                                             conf_level=FMT['nap_confidence'])
            else:
                logger.warning(f'Cannot calculate NAP for {nap_sel}')
        return cls(population_mean=pop_mean,
                   population_individual=pop_ind,
                   subjects=subjects,
                   nap=nap)


# ------------------------------------- Secondary-level displays
class EmaDisplay:
    """Container for all scenario and attribute displays for ONE (Sub-)Population
    either random-individual or mean,
    OR for ONE subject.
    """
    def __init__(self, scenarios, attributes):
        """
        :param scenarios: dict with (scenario_tuple, profile), where
            profile is a Profile instance for the selected scenario_tuple
        :param attributes: dict with (attr_effect, profile), where
            profile is an Profile instance for the selected attr_effect
        """
        self.scenarios = scenarios
        self.attributes = attributes

    def save(self, path):
        """Save all stored display objects in specified (sub-)tree
        :param path: path to directory for saving
        :return: None
        """
        if len(self.scenarios) > 0:
            for (d, sc_display) in self.scenarios.items():
                sc_display.save(path / 'scenarios')
        if len(self.attributes) > 0:
            for (d, a_display) in self.attributes.items():
                a_display.save(path / 'attributes')

    @classmethod
    def display(cls, m_xi):
        """
        :param m_xi: probability model for ema_model parameter vector xi
            either for population-individual, population-mean, or individual subject
        :return: a cls instance
        """
        xi = m_xi.rvs()
        scenarios = {sc: ScenarioProfile.display(xi, m_xi, sc)  # *** m_xi -> base
                     for sc in FMT['scenarios']}
        attributes = {a_effect: AttributeProfile.display(xi, m_xi, a_effect)  # *** m_xi -> base
                      for a_effect in FMT['attributes']}
        return cls(scenarios, attributes)


# ----------------------------------------------------------------------
class ScenarioProfile(Profile):
    """Container for all displays of ONE scenario effect
    in ONE (Sub-)Population, OR for ONE respondent.
    NOTE: Scenario profiles are displayed as
    CONDITIONAL probability of categories in ONE scenario dimension,
    GIVEN categories in other dimension(s), if requested.
    """
    @classmethod
    def display(cls, xi, m_xi, sc_keys):
        """Generate a probability-profile display for selected distribution and factor
        :param xi: 2D array of parameter-vector samples drawn from m_xi
        :param m_xi: a population or individual model instance
        :param sc_keys: tuple of one or more key(s) selected from emf.scenarios.keys()
        :return: single cls instance showing CONDITIONAL probabilities
            for sc_keys[0], GIVEN each combination (j1,..., jD) for sc_keys[1], ...
        """
        emf = m_xi.base.emf
        u = m_xi.base.scenario_prob(xi)
        u /= u.shape[1]  # = JOINT prob mass, incl. TestStage: sum(u) == 1
        u = _aggregate_scenario_prob(u, emf, sc_keys)
        # Now with reordered and aggregated joint probabilities
        # for selected subset of scenario dimensions, such that
        # u[s, j0, j1, ...] = s-th sample of joint probability for
        # emf.scenario[sc_keys[0]][j0], emf.scenario[sc_keys[1]][j1], ... etc.
        # u.shape == (u.shape[0], *emf.scenario_shape[emf.scenario_axes(sc_keys)])
        if len(sc_keys) > 1:
            s = np.sum(u, axis=1, keepdims=True)
            too_small = s <= 0
            n_underflow = np.sum(s <= 0.)
            if n_underflow > 0:
                logger.warning(f'Scenario display: {n_underflow} prob. sample(s) == 0, '
                               + f'with {sc_keys}. '
                               + 'Should not happen! Maybe too few EMA records?')
                s = np.maximum(s, np.finfo(float).tiny)
            u /= s
            # u[:, i, ...] = samples for CONDITIONAL probability of i-th category in sc_keys[0],
            # GIVEN ...-th category product of sc_keys[1:]
            u = u.reshape((*u.shape[:2], -1))
            # now linear-indexed in 3D with
            # u[:, :, j] = prob given j-th product of sc_keys[2:], if any given
            # case_labels[2] = all sc_keys[2:] joined  # ****************** !!!!
        q = np.percentile(u, FMT['percentiles'], axis=0)
        perc = fmt.fig_percentiles(q,
                                   y_label=FMT['sc_probability'],
                                   case_labels=_case_labels(emf.scenarios, sc_keys)
                                   )
        # --------------------------------------- percentile table:
        table = fmt.tab_percentiles(q, perc=FMT['percentiles'],
                                    cdf=None,  # **** always > 0, not displayed
                                    y_label=FMT['sc_probability'],
                                    case_labels=_case_labels(emf.scenarios, sc_keys)
                                    )
        # ---------------------------------------- sc_keys differences
        # NOTE: Comparing CONDITIONAL probabilities of categories in FIRST sc_keys dimension,
        # GIVEN categories in other dimensions.
        d = cred_diff(u, diff_axis=1, sample_axis=0, case_axis=2,
                      p_lim=FMT['credibility_limit'])
        diff = fmt.tab_credible_diff(d,
                                     diff_labels=_product_tuples(emf.scenarios, sc_keys[0:1]),
                                     case_labels=_product_tuples(emf.scenarios, sc_keys[1:])
                                     )
        # ---------------------------------------------------------------------
        return cls(perc, table, diff)


class AttributeProfile(Profile):
    """Container for displays of ONE attribute, and scenario_effect(s),
    in ONE (Sub-)Population, OR for ONE respondent.

    NOTE: Latent-variable results are displayed for each Attribute,
    GIVEN Scenario categories in requested Scenario dimensions,
    averaged across all OTHER Scenario dimension,
    weighted by Scenario probabilities in those dimensions.
    """
    # **** allow several attributes in one display ? *****
    @classmethod
    def display(cls, xi, m_xi, a_effect):
        """Create displays for a single attribute and requested scenario effects
        :param xi: 2D array of parameter-vector samples drawn from m_xi
        :param m_xi: a population or individual model instance
        :param a_effect: tuple(attribute_key, sc_keys)
        :return: single cls instance with all displays
        """
        emf = m_xi.base.emf
        u = m_xi.base.scenario_prob(xi)
        u /= u.shape[1]  # = JOINT prob mass, incl. TestStage: sum(u) == 1
        (a, sc) = a_effect
        # a = attribute key, sc_keys = tuple of scenario keys
        # file_name = a + '_vs_' + '*'.join(sc)
        theta = m_xi.base.attribute_theta(xi, a)
        # theta[s, k0, k1,...] = s-th sample of attribute a, given (k0, k1,...)-th scenario
        theta = _aggregate_scenario_theta(theta, u, emf, sc)
        # theta[s, j0, j1,...] = s-th sample of attribute a,
        #   given (scenarios[sc_keys[0]][j0], scenarios[sc_keys[1]][j1], ...)
        tau = np.median(m_xi.base.attribute_tau(xi, a), axis=0)
        # tau[l] = l-th median rating threshold for attribute a
        q = np.percentile(theta, FMT['percentiles'], axis=0)
        perc = fmt.fig_percentiles(q,
                                   y_label=a + ' (' + FMT['scale_unit'] + ')',
                                   case_labels=_case_labels(emf.scenarios, sc),
                                   cat_limits=tau,
                                   file_label=a)
        # --------------------------------------- percentile table:
        table = fmt.tab_percentiles(q, perc=FMT['percentiles'],
                                    cdf=_sample_cdf_0(theta, axis=0),
                                    y_label=a,
                                    case_labels=_case_labels(emf.scenarios, sc),
                                    file_label=a)
        # ---------------------------------------- attr differences
        # NOTE: comparing all scenario-categories, in all requested sc dimensions
        th_1 = theta.reshape((theta.shape[0], -1))
        # all sc_keys dimensions flattened into th_1[:, 1]
        d = cred_diff(th_1, diff_axis=1, sample_axis=0,
                      p_lim=FMT['credibility_limit'])
        diff = fmt.tab_credible_diff(d,
                                     y_label=a,  # for table header
                                     # diff_head=sc,  # not needed
                                     diff_labels=_product_tuples(emf.scenarios, sc),
                                     file_label=a)
        # ---------------------------------------------------------------------
        return cls(perc, table, diff)


# --------------------------------- classes for differences between groups
class GroupEffectSet:
    """Container for displays of differences between populations,
    as represented by subjects in separate groups
    """
    def __init__(self, population_mean=None, population_individual=None):
        """
        :param population_mean: dict with elements (a_label, AttributeGroupEffects instance)
        :param population_individual: dict with elements (a_label, AttributeGroupEffects instance)
            both with results separated by groups
        """
        self.population_mean = population_mean
        self.population_individual = population_individual

    def save(self, path):
        """Save all stored display objects in their corresponding sub-trees
        """
        if self.population_mean is not None:
            self.population_mean.save(path / 'population_mean')
        if self.population_individual is not None:
            self.population_individual.save(path / 'population_individual')

    @classmethod
    def display(cls, emm):
        """Generate all displays for ONE group
        :param emm: ema_model.EmaModel instance with several groups
        :return: cls instance with all displays of differences
            between predictive distributions for
            population random individual, AND/OR
            population mean
        """
        pop_ind = None
        pop_mean = None
        if FMT['population_individual']:
            pop_ind = EmaGroupDiff.display(emm,
                                           {g: emm_g.predictive_population_ind()
                                            for (g, emm_g) in emm.groups.items()})
        if FMT['population_mean']:
            pop_mean = EmaGroupDiff.display(emm,
                                            {g: emm_g.predictive_population_ind()
                                             for (g, emm_g) in emm.groups.items()})
        return cls(population_mean=pop_mean,
                   population_individual=pop_ind)


class EmaGroupDiff:
    """Container for displays of differences between (Sub-)Populations
    represented by separate ema_model.EmaGroupModel instances.
    """
    def __init__(self, scenarios, attributes):
        """
        :param scenarios: dict with (scenario_tuple, diff_profile), where
            profile is a Profile instance for the selected scenario_tuple
        :param attributes: dict with (attr_effect, diff_profile), where
            profile is an Profile instance for the selected attr_effect
        """
        self.scenarios = scenarios
        self.attributes = attributes

    def save(self, path):
        """Save all stored display objects in specified (sub-)tree
        """
        if len(self.scenarios) > 0:
            for (d, sc_display) in self.scenarios.items():
                sc_display.save(path / 'scenarios')
        if len(self.attributes) > 0:
            for (d, a_display) in self.attributes.items():
                a_display.save(path / 'attributes')

    @classmethod
    def display(cls, emm, groups):
        """
        :param emm: ema_model.EmaModel instance with several groups
        :param groups: dict with elements (g_id, g_model), where
            g_id is a tuple of one or more tuple(group_factor, factor_category),
            g_model is a predictive ema_model.PredictivePopulationModel instance
            i.e. NOT emm.groups
        :return: single cls instance
        """
        xi = [g_model.rvs()
              for g_model in groups.values()]
        # xi[g][s, :] = s-th sample of parameter vector for g-th group
        scenarios = {sc: ScenarioDiff.display(xi, emm, sc)
                     for sc in FMT['scenarios']}
        attributes = {a_effect: AttributeDiff.display(xi, emm, a_effect)
                      for a_effect in FMT['attributes']}
        return cls(scenarios, attributes)


class ScenarioDiff(Profile):
    """Container for all displays of group differences in ONE scenario effect
    """
    @classmethod
    def display(cls, xi, emm, sc_keys):
        """Generate a probability-profile display for selected distribution and factor
        :param xi: list of 2D arrays of parameter-vector samples
            len(xi) == len(emm.groups)
        :param emm: ema_model.EmaModel object
        :param sc_keys: tuple of one or more key(s) selected from emf.scenarios.keys()
        :return: single cls instance
        """
        emf = emm.base.emf
        u = [emm.base.scenario_prob(xi_g) for xi_g in xi]
        # u[g][s, k0, k1, ...] = s-th sample of (k0, k1,...)-th scenario prob in g-th population
        for u_g in u:
            u_g /= u_g.shape[1]  # = JOINT prob mass, incl. k0 = TestStage index:
            # sum(u_g) == 1
        u = [_aggregate_scenario_prob(u_g, emf, sc_keys)
             for u_g in u]
        # Now with reordered and aggregated joint probabilities
        # for selected subset of scenario dimensions, such that
        # u[g, s, j0, j1, ...] = s-th sample of joint probability for
        # emf.scenario[sc_keys[0]][j0], emf.scenario[sc_keys[1]][j1], ... etc.
        # file_name = '_'.join(sc_keys)
        if len(sc_keys) > 1:
            u_check = []
            for u_g in u:
                s = np.sum(u_g, axis=1, keepdims=True)
                n_underflow = np.sum(s <= 0.)
                if n_underflow > 0:
                    logger.warning(f'ScenarioDiff display: {n_underflow} prob. sample(s) == 0. '
                                   + 'Should not happen! Maybe too few responses?')
                    s = np.maximum(s, np.finfo(float).tiny)
                u_check.append(u_g / s)
            u = u_check
            # u = [u_g / np.sum(u_g, axis=1, keepdims=True)
            #      for u_g in u]
            # u[g][s, i, ...] = samples for CONDITIONAL probability of i-th category in sc_keys[0],
            # GIVEN ...-th category product of sc_keys[1:]
            u = [u_g.reshape((u_g.shape[0], -1))
                 for u_g in u]
            # now each u_g linear-indexed in 2D with
            # u[g][s, j] = j-th product of sc_keys
        # ---------------------------------------- sc_keys differences
        # NOTE: Comparing CONDITIONAL probabilities of categories in FIRST sc_keys dimension,
        # GIVEN categories in other dimensions.
        d = cred_group_diff(u, sample_axis=0, case_axis=1,
                            p_lim=FMT['credibility_limit'])
        diff = fmt.tab_credible_diff(d,
                                     diff_labels=emm.groups.keys(),
                                     case_labels=_product_tuples(emf.scenarios, sc_keys)
                                     )
        # ---------------------------------------------------------------------
        return cls(diff=diff)


class AttributeDiff(Profile):
    """Container for all displays of group differences in ONE Attribute-by-Scenario effect
    """
    # **** allow several attributes in one display ? ***************
    @classmethod
    def display(cls, xi, emm, a_effect):
        """Create displays for a single attribute and requested scenario effects
        :param xi: list of 2D arrays of parameter-vector samples
            len(xi) == len(emm.groups)
        :param a_effect: tuple (attr_key, sc_keys), where
            sc_keys is a tuple of one or more key(s) selected from emf.scenarios.keys()
        :param emm: ema_model.EmaModel object
        :return: single cls instance
        """
        emf = emm.base.emf
        u = [emm.base.scenario_prob(xi_g) for xi_g in xi]
        # u[g][s, k0, k1, ...] = s-th sample of (k0, k1,...)-th scenario prob in g-th population
        for u_g in u:
            u_g /= u_g.shape[1]  # = JOINT prob mass, incl. k0 = TestStage index:
            # sum(u_g) == 1
        (a, sc) = a_effect
        # a = attribute key, sc_keys = tuple of scenario keys
        # file_name = a + '_vs_' + '*'.join(sc)
        theta = [emm.base.attribute_theta(xi_g, a)
                 for xi_g in xi]
        # theta[g][s, k0, k1,...] = s-th sample of attribute a, given (k0, k1,...)-th scenario
        theta = [_aggregate_scenario_theta(theta_g, u_g, emf, sc)
                 for (theta_g, u_g) in zip(theta, u)]
        # theta[s, j0, j1,...] = s-th sample of attribute a,
        #   given (scenarios[sc_keys[0]][j0], scenarios[sc_keys[1]][j1], ...)
        #   averaged across OTHER scenarios, weighted by scenario-probabilities.
        # ---------------------------------------- attr differences
        # NOTE: comparing all scenario-categories, in all requested sc dimensions
        th_1 = [theta_g.reshape((theta_g.shape[0], -1))
                for theta_g in theta]
        # all sc_keys dimensions flattened into th_1[g][s, :]
        d = cred_group_diff(th_1, sample_axis=0, case_axis=1,
                            p_lim=FMT['credibility_limit'])
        diff = fmt.tab_credible_diff(d,
                                     y_label=a,
                                     file_label=a,
                                     diff_labels=emm.groups.keys(),
                                     case_labels=_product_tuples(emf.scenarios, sc)
                                     )
        # ---------------------------------------------------------------------
        return cls(diff=diff)


# ---------------------------------- Help functions:
def _aggregate_scenario_prob(u, emf, sc):
    """Aggregate probability-mass samples to keep only selected factor axes
    :param u: multi-dim array with probability-mass samples
        u[s, k0, k1,...] = s-th sample of JOINT prob
            for (k0, k1,...)-th scenario, as defined by emf.scenarios.items()
        u.shape == (n_samples, *emf.scenario_shape)
    :param emf: ema_data.EmaFrame instance for model generating u
    :param sc: tuple with selected scenario keys to display
    :return: array uf with aggregated joint probabilities
        uf[s, j0, j1, ...] = s-th sample of joint probability for
        scenario[sc[0]][j0], scenario[sc[1]][j1], ... etc.
        uf.shape == (u.shape[0], *emf.scenario_shape[emf.scenario_axes(sc)])
    """
    axes = tuple(1 + i for i in emf.scenario_axes(sc))
    uf = np.moveaxis(u, axes, tuple(range(1, 1+len(axes))))
    # with desired sc_keys axes first after 0
    sum_axes = tuple(range(1+len(axes), uf.ndim))
    uf = np.sum(uf, axis=sum_axes)
    # summed across all OTHER axes, except those in sc_keys
    return uf


def _aggregate_scenario_theta(th, u, emf, sc):
    """Aggregate attribute location sample arrays to keep only selected factor axes
    :param th: multi-dim array with attribute-location samples
        th[s, k0, k1,...] = s-th sample of latent variable theta
            in (k0, k1,...)-th scenario, as defined by emf.scenarios.items()
        th.shape == (n_samples, *emf.scenario_shape)
    :param u: multi-dim array with corresponding samples of normalized scenario probabilities
        u[s, k0, k1, ...] = s-th sample of normalized probability of (k0, k1,...)-th scenario
        sum_(k0, k1, ...) u[s, k0, k1, ...] == 1, for all s
        u.shape == th.shape
    :param emf: ema_data.EmaFrame instance for model generating u
    :param sc: tuple with selected scenario keys to display
    :return: th_a = array with aggregated attribute locations
        th_a[s, j0, j1, ...] = s-th sample of conditional attribute location,
        GIVEN scenario[sc[0]][j0], scenario[sc[1]][j1], ... etc.
        averaged across all OTHER Scenarios not included in sc.
        th_a.shape == (n_samples, *emf.scenario_shape[emf.scenario_axes(sc)])

    2021-12-17, probability-averaged across non-included Scenarios
    """
    keep_axes = tuple(1 + i for i in emf.scenario_axes(sc))
    w = np.sum(u, axis=keep_axes, keepdims=True)
    # = normalized scenario prob, for every sample
    th = np.moveaxis(th, keep_axes, tuple(range(1, 1 + len(keep_axes))))
    w = np.moveaxis(w, keep_axes, tuple(range(1, 1 + len(keep_axes))))
    # th and w now with desired sc_keys axes first after sample-axis = 0
    aggregate_axes = tuple(range(1+len(keep_axes), th.ndim))
    # = all OTHER axes, except those in sc_keys
    # th_old = np.mean(th, axis=aggregate_axes)
    # # averaged across all aggregate_axes. version <= 0.6
    th = np.sum(th * w, axis=aggregate_axes)
    # = probability-averaged across all aggregate_axes. version >= 0.7
    return th


def _case_labels(label_dict, key_list):
    """Sequence of case tuples
    :param label_dict: a dict with elements (factor_key, label_list)
    :param key_list: sequence of keys to label_dict
    :return: labels = list of tuples, with
        i-th tuple = (keys[i], label_dict[keys[i])
    """
    # **** return dict instead ??? **************
    return [(gf, label_dict[gf]) for gf in key_list]


def _product_labels(label_dict, key_list):
    """Iterator of tuples, each
    a product of one category from each desired factor dimension
    :param label_dict: a dict with elements (factor_key, label_list)
    :param key_list: list of keys to label_dict
    :return: labels = iterator yielding tuples, with
        ...-th tuple = (label_dict[keys[0][i0], ..., label_dict[keys[D][iD])
        where D = len(keys)
        with last index iD varying fastest, i0 slowest, in the product tuples
    """
    return product(*(label_dict[gf] for gf in key_list))


def _product_tuples(label_dict, keys):
    """Iterator of tuples, each
    a product of one category from each desired factor dimension
    :param label_dict: a dict with elements (factor_key, label_list)
    :param keys: list of keys to label_dict
    :return: labels = iterator yielding tuples, with
        ...-th tuple = ((keys[0]: label_dict[key_list[0][i0]),
                        , (...,), (keys[D]: label_dict[key_list[D][iD])}
        where D = len(key_list)
        with last index iD varying fastest, i0 slowest, in the product tuples
    """
    return product(*(product([gf], label_dict[gf]) for gf in keys))


# def _make_subject_table(a_label, a_model):
#     """Tabulate all quality estimates for all subjects in ONE group, for ONE attribute
#     :param a_label: attribute string label
#     :param a_model: pc_result.PairedCompGroupModel instance
#     :return: a TableRef instance with results for all subjects
#     """
#     pcf = a_model.pcf
#     q = [s_model.quality_samples[..., 1:]
#          for s_model in a_model.subjects.values()]
#     # q[s][n, t, :] = n-th sample vector for s-th subject in t-th test condition
#     # skipping first reference object
#     q_perc = np.array([np.percentile(q_s,
#                                      FMT['percentiles'],
#                                      axis=0)
#                        for q_s in q])
#     # q_perc[s, p, t, i] = p-th percentile of i-th object for s-th subject in t-th test condition
#     cdf = np.array([_sample_cdf_0(q_s) for q_s in q])
#     # cdf[s, t, :] = P[q<= 0] for s-th subject in t-th test condition
#     # *** sort subjects by labels alphabetically
#     s_list = [*a_model.subjects.keys()]
#     s_index = np.argsort(s_list)
#     q_perc = q_perc[s_index, ...]
#     cdf = cdf[s_index, ...]
#     s_list.sort()
#
#     q_perc = q_perc.transpose((1, 0, 2, 3))  # **** for current tab_percentiles
#     # q_perc[p, s, t, i] = n-th sample of i-th object for s-th subject in t-th test condition
#     return fmt.tab_percentiles(q_perc,
#                                cdf=cdf,
#                                perc=FMT['percentiles'],
#                                y_label=a_label,
#                                case_labels=[(FMT['subject'], s_list),
#                                             *pcf.test_factors.items(),
#                                             (FMT['object'], pcf.objects_disp[1:])],
#                                case_order=[FMT['subject'],
#                                            FMT['object'],
#                                            *pcf.test_factors.keys()]
#                                )


def _dir_name(g, sep='_'):
    """make sure group name is a possible directory name
    :param g: string or tuple of strings
    :return: string to be used as directory
    """
    if type(g) is tuple:  # several strings
        g = sep.join(_dir_name(g_s, sep='_')
                     for g_s in g)
    return g


def _sample_cdf_0(u, axis=0):
    """
    Probability that U <= 0 calculated from set of samples.
    :param u: array of samples drawn from U
    :param axis: (optional) axis with independent samples
        e.g., mean of U == np.mean(u, axis=axis)
    :return: array p with
        p[...] = P(U[...] <= 0)
        p.shape == u.shape with axis removed
    """
    n = u.shape[axis]
    return (np.sum(u <= 0, axis=axis) + 0.5) / (n + 1)


def fig_comments():
    """Generate figure explanations.
    :return: comment string
    """
    p_min = np.amin(FMT['percentiles'])
    p_max = np.amax(FMT['percentiles'])
    c = f"""Figure Explanations:
    
Figure files with names like
someScenarioName_xxx.pdf, someAttributeName_xxx.pdf or xxx.jpg, or similar,
display user-requested percentiles (markers) and credible intervals (vertical bars) 
The vertical bars show the range between {p_min:.1f}- and {p_max:.1f}- percentiles.

Median ordinal-response thresholds for perceptual Attributes, if requested,
are indicated by thin lines extending horizontally across the graph.

The credible ranges include all uncertainty
caused both by real inter-individual perceptual differences
and by the limited number of responses by each listener.
"""
    return c


def table_comments():
    c = """Table Explanations:

*** Tables of Percentiles:
Files with names like someScenarioName_xxx.tex, someAttributeName_xxx.tex or *.txt
show numerical versions of the information in corresponding percentile plots.
Percentiles, credible ranges, and marginal probabilities for negative and positive values are shown.

*** Tables of Credible Differences:
Files with names like someAttribute-diff_xxx.tex or *.txt 
show a list of Attribute (in Scenario) pairs
which are ALL JOINTLY credibly different
with the tabulated credibility.
The credibility value in each row is the JOINT probability
for the pairs in the same row and all rows above it.
Thus, the joint probability values already account for multiple comparisons,
so no further adjustments are needed.
"""
    if len(FMT['nap']) > 0:
        c += """\n*** NAP Tables:
Table files with names like someAttribute_sc-dim_xxx.tex or *.txt
show estimated NAP effect(s) of a Scenario Dimension "sc_dim" with exactly TWO categories.
NAP values > 0.5 indicate that the SECOND category is rated higher than the FIRST.
"""
        cl = FMT['nap_confidence']
        c += f'{cl:.1%} Confidence Intervals were estimated by the "MW-N" method (Feng et al., 2017).\n'
    return c



"""This module defines classes to access and store recorded EMA data,
and methods and functions to read and write such data.
Data may (*future*) be stored in various file formats,
for now only Excel workbook (.xlsx) files can be used.

Each EMA Record is a list of nominal and ordinal elements, defining
* a subject ID label,
* a single SCENARIO specified by selected categories in ONE or more Scenario Dimension(s),
* ordinal RATING(s) for ZERO, ONE, or MORE perceptual ATTRIBUTE(s) in the current scenario.

*** Class Overview:

EmaFrame: defines layout and category labels of data in each EMA Record.
    Some properties of an EmaFrame instance can define selection criteria
    for a subset of data to be included for analysis.

EmaDataSet: all data to be used as input for statistical analysis.

*** Input File Formats:

* xlsx: Data can (for now) be imported ONLY from Excel workbook (xlsx) files,
with response data stored in designated columns of one or more worksheets.
Files in this format must be saved with extension '.xlsx'.

(Other data file formats may be allowed in future versions.)

*** Input Data Files:

All input files from an experiment must be stored in ONE directory tree.

If results are to be analyzed for more than one GROUP of test subjects,
the data for each group must be stored in separate sub-directories
within the specified top directory.

Groups are identified by a tuple of grouping pairs (group_factor, group_category), where
group_factor is a grouping dimension, e.g., 'Age', or 'Gender', and
group_category is a nominal label within the factor, e.g., 'old', or 'female'.

A sequence of one element from each grouping pair must define a unique path
to files including data for subjects in ONE group.
Group directories must have names like, e.g., 'Age_old' for group = ('Age', 'old')

Subject file names are arbitrary, although they may be somehow associated with
the encoded name of the participant, to facilitate data organisation.

Each data file in ONE group directory may contain EMA records
from ONE or SEVERAL subjects in this group.
Each subject in the same group must have a unique subject ID.
The subject ID must be given either in the sheet title
or on each row in a specified sheet column.

Different files in the same sub-directory
may include data for the same subject in the group,
e.g., results obtained in different test stages,
or simply for replicated EMA responses from the same subject.

Subjects in different groups may have the same subject ID values,
because the groups are separated anyway,
but normally the subject IDs should be unique across all groups.

*** Example Directory Tree:

Assume we have data files in the following directory structure:
~/ema_study / Age_old / Gender_male, containing files Data_EMA_64.xlsx and Response_Data_LAB_64.xlsx
~/ema_study / Age_old / Gender_female, containing files Subjects_EMA_64.xlsx and Data_EMA_65.xlsx
~/ema_study / Age_young / Gender_male,  containing files EMA_64.xlsx and EMA_65.xlsx
~/ema_study / Age_young / Gender_female, containing files EMA_64.xlsx and EMA_65.xlsx

Four separate groups may then be defined by factors Age and Gender,
and the analysis may be restricted to data in files with names including 'EMA_64'.


*** Accessing Input Data for Analysis:
*1: Create an EmaFrame object defining the experimental layout, e.g., as:

emf = EmaFrame(scenarios={'CoSS': [f'{i}.' for i in range(1, 8)],
                          'Important': ('Slightly', 'Medium', 'Very'),
                          },  # nominal variables
               attribute_grades={'Speech': ('Very Hard', 'Fairly Hard', 'Fairly Easy','Very Easy')},
        )
NOTE: Letter CASE is always distinctive, i.e., 'Male' and 'male' are different categories.

*2: Load all test results into an EmaDataSet object:

ds = EmaDataSet.load(emf, path='~/ema_study',
                    grouping={'Age': ('young', 'old'),
                              'Gender': ('female', 'male'),
                              'Test': ('EMA_64',)}
                    fmt='xlsx',
                    subject='sheet',    # xlsx sheet title is subject ID
                    ema_vars={'CoSS': 'B',       # column B contains Scenario 'CoSS' category
                              'Important': 'C',  # column C contains Scenario 'Important' category
                              'Speech': 'D'     # column D contains 'Speech' ordinal rating
                              }
                    )

The loaded data structure ds can then be used as input for analysis.
The parameter emf is a EmaFrame object that defines the variables to be analyzed.

*** Selecting Subsets of Data for Analysis:
It is possible to define a data set including only a subset of recorded data files.

For example, assume we want to analyze only two groups, old males, and old females.
and only responses for Scenario dimension 'CoSS'.
Then we must define a new EmaFrame object, and load only a subset of group data:

emf = EmaFrame(scenarios={'CoSS': [f'{i}.' for i in range(1, 8)],
                          },  # nominal variables
               attribute_grades={'Speech': ('Very Hard', 'Fairly Hard', 'Fairly Easy','Very Easy')},
        )
ds = EmaDataSet.load(emf, path='~/ema_study',
                    grouping={'Age': ('old',),
                              'Gender': ('female', 'male'),
                              'Test': ('EMA_64',)}
                    fmt='xlsx',
                    subject='sheet',    # xlsx sheet title is subject ID
                    ema_vars={'CoSS': 'B',       # column B contains Scenario 'CoSS' category
                              'Speech': 'D'     # column D contains 'Speech' ordinal rating
                              }
                    )

*** Version History:
* Version 0.9?:
2022-02-xx, allow missing response for some Attribute(s) ?!

* Version 0.8.3:
2022-03-08, minor fix for FileReadError error message

* Version 0.8.1:
2021-02-27, fix EmaDataSet.load(), _gen_group_file_paths(), _groups(), for case NO grouping

* Version 0.5.1:
2021-11-26, EmaDataSet.load warning for input argument problems

* Version 0.5:
2021-10-15, first functional version
2021-11-18, grouping moved from EmaFrame -> EmaDataSet.load
2021-11-20, EmaDataSet.ensure_complete
2021-11-23, Group dir name MUST include both (g_factor, g_cat), e.g., 'Age_old'
2021-11-xx, allow empty attribute_grades
"""
# ******** check file exception message output ********************
# *** EmaDataSet.initialize + add method ? load = initialize + add
# *** allow Pandas input ?
import numpy as np
from pathlib import Path
from itertools import product
# import json
import logging

from EmaCalc import ema_file_xlsx
from EmaCalc.ema_file import FileReadError

FILE_CLASS = {'xlsx': ema_file_xlsx.EmaFile,
              }
# = mapping of file format suffix to file reader module, only ONE format for now...

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# ------------------------------------------------------------------
class EmaFrame:
    """Defines structure of EMA recorded dataset for one analysis.
    """
    def __init__(self,
                 scenarios=None,
                 stage_key='Stage',  # default scenario key for test stage
                 attribute_grades=None,
                 ):
        """
        :param scenarios: (optional) dict or iterable with elements (scene_factor, category_list), where
            scene_factor is a string label identifying one scenario "dimension",
            category_list is a list of labels for NOMINAL categories within this scene_factor.
        :param attribute_grades: (optional) dict or iterable with elements (attribute, response_list),
            attribute is string id of a rated perceptual attribute,
            response_list is an iterable with ORDINAL categories, strings or numeric
        :param stage_key: (optional) string key, MUST be member of scenarios.keys()
            scenarios[stage_key] = list of test stages (e.g., before vs after treatment),
                specified by experimenter, i.e., NOT given as an EMA response

        NOTE: scenarios, attribute_grades, grouping may define a subset of
            data in input data files, if not all variant are to be analyzed.

            Category label strings may be shorter than those occurring in data files,
            and will be considered matching if they agree within the given length.
        """
        self.stage_key = stage_key
        if scenarios is None:
            scenarios = dict()
        else:
            scenarios = dict(scenarios)
        # ensure first scenario key == self.stage_key:
        # **** requires python >= 3.9
        stage_dict = {self.stage_key: scenarios.pop(self.stage_key, (None,))}
        self.scenarios = stage_dict | scenarios

        if attribute_grades is None:
            self.attribute_grades = dict()
        else:
            self.attribute_grades = dict(attribute_grades)

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t\t' +
                ',\n\t\t'.join(f'{key}={repr(v)}'
                               for (key, v) in vars(self).items()) +
                '\n\t\t)')

    @property
    def scenario_shape(self):
        """tuple with number of nominal categories for each scenario factor"""
        return tuple(len(sc_list)
                     for sc_list in self.scenarios.values())

    @property
    def n_scenarios(self):  # **** needed ?
        return np.prod(self.scenario_shape, dtype=int)

    def scenario_axes(self, sc):
        """
        :param sc: sequence of one or more scenario keys
        :return: tuple of corresponding numerical axes
        """
        sc_keys = list(self.scenarios.keys())
        sc_ind = []
        for sc_i in sc:
            try:
                sc_ind.append(sc_keys.index(sc_i))
            except ValueError:
                logger.warning(f'{repr(sc_i)} is not a scenario key')
        return sc_ind

    @property
    def rating_shape(self):
        """tuple with number of ordinal response levels, for each attribute
        """
        return tuple(len(r_list)
                     for r_list in self.attribute_grades.values())

    @property
    def n_stages(self):
        # == scenario_shape[0]
        return len(self.scenarios[self.stage_key])

    def accept(self, ema):
        """Check that ema_record includes acceptable values
         for all variables required in self
        :param ema: dict with elements (key, value) from observed EMA record
        :return: boolean = True if acceptable
        """
        # *** future: allow missing data for some attribute(s) ?
        try:
            scenario_ok = all(0 <= _index_matching(sc_list, ema[sc_key])
                              for (sc_key, sc_list) in self.scenarios.items())
            rating_ok = all(0 <= _index_matching(r_list, ema[r_key])
                            for (r_key, r_list) in self.attribute_grades.items())
            return scenario_ok and rating_ok
        except (KeyError, ValueError):
            return False

    def scenario_index(self, r):
        """Encode an EMA record into an index tuple identifying the scenario
        :param r: dict with (key, value) pairs of scenarios and attribute_grades
        :return: sc_i = tuple (k_0, k_1, k_2, ...), where
            k_0 = test-stage index,
            k_i = index of i-th scenario
            len(sc_i) == len(self.scenarios)
        """
        try:
            return tuple(_index_matching(sc_list, r[sc_key])
                         for (sc_key, sc_list) in self.scenarios.items())
        except (KeyError, ValueError):
            return None

    def rating_index(self, r):
        """Encode an EMA record into an index tuple identifying rating responses
        :param r: dict with (key, value) pairs of scenarios and attribute_grades
        :return: y = tuple, with
            y[i] = scalar integer = ordinal value of i-th rating response
            y[i] = None, if response is missing
            len(y) == len(self.attribute_grades)
        """
        try:
            return tuple(_index_matching(r_list, r[r_key])
                         for (r_key, r_list) in self.attribute_grades.items())
        except (KeyError, ValueError):
            return None


# ------------------------------------------------------------
class EmaDataSet:
    """Container of all data input for one complete EMA study.
    """
    def __init__(self, emf, groups):
        """
        :param emf: an EmaFrame instance
        :param groups: dict with elements (group_id: group_dict), where
            group_id = tuple with one or more pairs (g_factor, g_category),
                identifying the sub-population,
            group_dict = dict with elements (subject, ema_list), where
            ema_list = list with elements ema, where
                ema = tuple with elements (key, value), where
                key is one of [stage_key, scene_factor, attribute],
                value is one allowed category for this key.
                len(ema) = emf.n_variables
        """
        self.emf = emf
        self.groups = groups

    def __repr__(self):
        def sum_n_records(g_subjects):
            """Total number of EMA records across all subjects in group"""
            return sum(len(s_ema) for s_ema in g_subjects.values())
        # ---------------------------------------------------------------
        return (self.__class__.__name__ + '(\n\t'
                + f'emf= {self.emf},\n\t'
                + 'groups= {' + '\n\t\t'
                + '\n\t\t'.join((f'{g}: {len(g_subjects)} subjects '
                                 + f'with {sum_n_records(g_subjects)} EMA records in total,')
                                for (g, g_subjects) in self.groups.items())
                + '\n\t\t})')

    @classmethod
    def load(cls, emf, path,
             subject,
             ema_vars,
             grouping=None,
             fmt='xlsx'):
        """Create one class instance with selected data from input files.
        :param emf: EmaFrame instance
        :param path: string or Path defining top of directory tree with all data files
        :param subject: address where to find subject ID in specific file object
        :param ema_vars: dict with args to specific file object,
            typically with elements (ema_key, column_address)
        :param grouping: (optional) dict or iterable with elements (group_dim, category_list),
            where
            group_dim is a string label identifying one "dimension" of sub-populations,
            category_list is a list of labels for allowed categories within group_factor.
            May be left empty, if only one (unnamed) group is included.
        :param fmt: (optional) string with file suffix for data files.
            If undefined, all files are tried, so (*future*) mixed file formats can be used as input.
        :return: a single cls object
        """
        assert (fmt in FILE_CLASS.keys() or
                fmt in ['', None]), 'Unknown file format: ' + fmt
        # check that ema_vars.keys() are either Scenario or Attribute keys:
        emf_keys = set(emf.scenarios.keys()) | set(emf.attribute_grades.keys())
        ema_keys = set(ema_vars.keys())
        missing_emf = emf_keys - ema_keys
        if len(missing_emf) > 0:
            logger.warning(f'No file address defined for {missing_emf}')
        missing_ema = ema_keys - emf_keys
        if len(missing_ema) > 0:
            logger.warning(f'File address(es) {missing_ema} defined, but not needed.')
        path = Path(path)
        if grouping is None:
            grouping = dict()
        else:
            grouping = dict(grouping)
        groups = {g: dict() for g in _groups(grouping)}
        # = empty dict for subjects in each group
        # **** up to here -> classmethod initialize
        # **** following: -> add method, to allow collecting data from different file formats
        for (g, g_path) in _gen_group_file_paths(path, fmt, [*grouping.items()]):
            logger.info(f'Reading {g_path}')
            ema_file = FILE_CLASS[fmt](g_path,
                                       subject=subject,
                                       ema_vars=ema_vars)
            try:
                for r in ema_file:
                    # add stage if only one allowed and not specified
                    if emf.n_stages == 1 and emf.stage_key not in r.ema.keys():
                        r.ema[emf.stage_key] = emf.scenarios[emf.stage_key][0]
                    if emf.accept(r.ema):
                        if r.subject not in groups[g]:
                            groups[g][r.subject] = []
                            # = initialized empty ema list for this new subject
                        groups[g][r.subject].append(r.ema)
                    else:
                        logger.debug(f'Not Used: Subject {r.subject}: {r.ema}')
            except FileReadError as e:
                logger.warning(e)  # and just try next file
        return cls(emf, groups)

    # def add method, to include data from new files with different layout ???

    def save(self, dir, subject, ema_vars, allow_over_write=False, fmt='xlsx'):
        """Save self.groups in a directory tree with one folder for each group,
        with one file for each subject.
        :param dir: Path or string defining the top directory where files are saved
        :param allow_over_write: boolean switch, over-write files if True
        :param fmt: string label specifying file format
        :param subject: string to identify subject in specific file object,
        :param ema_vars: dict with args to specific file object,
            typically with elements (ema_key, column_address)
        :return: None
        """
        for key in self.emf.scenarios.keys():
            if key not in ema_vars.keys():
                logger.warning('No column address for Scenario ' + repr(key) + ': Not saved!')
        for key in self.emf.attribute_grades.keys():
            if key not in ema_vars.keys():
                logger.warning('No column address for Attribute ' + repr(key) + ': Not saved!')
        dir = Path(dir)
        for (g, group_data) in self.groups.items():
            g = _dir_name(g, '/')
            if len(g) == 0:
                g_path = dir
            else:
                g_path = dir / g
            g_path.mkdir(parents=True, exist_ok=True)
            for (s_id, s_items) in group_data.items():
                try:
                    p = (g_path / s_id).with_suffix('.' + fmt)  # one file per subject
                    s_file = FILE_CLASS[fmt](file_path=p,
                                             subject=subject,
                                             ema_vars=ema_vars)
                    s_file.save(subject=s_id, items=s_items,
                                allow_over_write=allow_over_write)
                except (KeyError, NotImplementedError):
                    raise RuntimeError(f'Cannot save {self.__class__.__name__} in {repr(fmt)} format')

    def ensure_complete(self):
        """Check that we have at least one subject in every sub-population category,
        with at least one ema record for each subject (already checked in load method).
        :return: None

        Result:
        self.groups may be reduced:
        subjects with no records are deleted,
        groups with no subjects are deleted
        logger warnings for missing data.
        """
        for (g, g_subjects) in self.groups.items():
            incomplete_subjects = set(s for (s, s_ema) in g_subjects.items()
                                     if len(s_ema) == 0)
            for s in incomplete_subjects:
                logger.warning(f'No EMA data for subject {repr(s)} in group {repr(g)}. Deleted!')
                del g_subjects[s]
        incomplete_groups = set(g for (g, g_subjects) in self.groups.items()
                                if len(g_subjects) == 0)
        for g in incomplete_groups:
            logger.warning(f'No subjects in group {repr(g)}. Deleted!')
            del self.groups[g]
        if len(self.groups) == 0:
            raise RuntimeError('No EMA data in any group.')


# -------------------------------------------- module help functions

def _dir_name(g, sep='_'):
    """make sure group name is a possible directory name
    :param g: string or tuple of strings
    :return: string to be used as directory
    """
    if type(g) is tuple:  # one or more (g_factor, g_cat) tuples
        g = sep.join(_dir_name(g_s, sep='_')
                     for g_s in g)
    return g


def _groups(group_factors):
    """Generate group labels from group_factors tree
    :param group_factors: dict or iterable with elements (group_factor, category_list),
    :return: generator of all combinations of (gf, gf_category) pairs from each group factor
        Generated pairs are sorted as in group_factors
    """
    if len(group_factors) == 0:  # NO grouping
        return [tuple()]  # ONE empty group label
    else:
        return product(*(product([gf], gf_cats)
                         for (gf, gf_cats) in group_factors.items())
                       )


def _gen_group_file_paths(path, fmt, group_factors, g_tuple=()):
    """Generator of group keys and corresponding file Paths, recursively, for all groups
    :param path: Path instance defining top directory to be searched
    :param fmt: file suffix of desired files
    :param group_factors: list of tuples (g_factor, labels)
    :param g_tuple: list of tuples (g_factor, factor_label),
        defining a combined group label or a beginning of a complete such label
    :return: generator of tuples (group_key, file_path), where
        group_key is an element of emf.groups,
        file_path is a Path object to a file that may hold count data for the group.
    """
    for f in path.iterdir():
        if len(group_factors) == 0:  # now at lowest grouping level in directory tree
            if f.is_file():  # include all files here and in sub-directories
                if fmt in f.suffix:
                    # print(f'Reading file {f}')
                    yield g_tuple, f
            elif f.is_dir():  # just search sub-tree recursively
                yield from _gen_group_file_paths(f, fmt,
                                                 group_factors,
                                                 g_tuple)
        else:  # len(grouping) >=1
            g_factor_key = group_factors[0][0]
            for g_cat in group_factors[0][1]:
                factor_cat = (g_factor_key, g_cat)
                # = new tuple to be included in final group label
                if f.is_dir():
                    # if f.name.find(g_cat) == 0:
                    if (f.name.find(g_factor_key) == 0
                            and f.name.find(g_cat) == len(g_factor_key) + 1):
                        # iterate recursively in sub-directory
                        yield from _gen_group_file_paths(f, fmt,
                                                         group_factors[1:],
                                                         (*g_tuple, factor_cat))
                elif f.is_file() and len(group_factors) == 1:
                    # at final sub-directory level, also accept group category in file name:
                    if (g_factor_key in f.name) and (g_cat in f.name) and fmt in f.suffix:
                        # print(f'Reading file {f}')
                        yield (*g_tuple, factor_cat), f


def _index_matching(category_list, label):
    """Like list.index, but allowing sub-string match
    :param category_list: list of string-valued labels
    :param label: a single label
    :return: index in str_list of element matching initial part of label
    """
    for (i, c_label) in enumerate(category_list):
        if label == c_label:
            return i
        elif label.find(c_label) == 0:
            return i
    raise ValueError('Label not matched')


# -------------------------------------------- TEST:
if __name__ == '__main__':
    # work_dir = Path('../..').resolve()
    home_dir = Path.home()
    work_dir = home_dir / 'Documents' / 'PythonTools' / 'EmaCalc_test'

    # = ORCA profile
    data_dir = work_dir / 'ORCA fake groups'
    print(f'*** Testing _gen_group_file_paths in {data_dir}')

    emf = EmaFrame(scenarios={'CoSS': [f'{i}.' for i in range(1, 8)],
                              'Viktigt': ['Lite / inte alls viktigt',
                                          'Ganska viktigt',
                                          'Mycket viktigt'],
                              # 'Test': (None, None),
                              },  # nominal variables
                   attribute_grades={'Speech': ['Mycket svårt',
                                                'Ganska svårt',
                                                'Lite / inte alls svårt']},  # ordinal variables
                   )
    print('emf=\n', emf)

    group_factors = {'Age': ['young', 'old'],
                     'Gender': ['male', 'female'],
                     'ORCA': ['AR_64']}
    # for (g, g_path) in _gen_group_file_paths(data_dir, 'xlsx', [*emf.grouping.items()]):
    for (g, g_path) in _gen_group_file_paths(data_dir, 'xlsx', [*group_factors.items()]):
        print('g= ', g, ': g_path= ', g_path)

    ds = EmaDataSet.load(emf, data_dir,
                         grouping=group_factors,
                         subject='A',
                         ema_vars={'CoSS': 'J',
                                   'Speech': 'P',
                                   'Viktigt': 'O'
                                   }
                         )
    print('ds= ', ds)

    # ***** TEST Empty scenarios or Empty attribute_grades *************'



"""This module estimates the effect measure Non-overlap of All Pairs (NAP),
to be used as a non-Bayesian measure of
difference between EMA Attribute Ratings.

This module is now part of package EmaCalc, but might also be distributed separately.

The NAP measure is a non-parametric estimate of the probability that
Attribute Rating X > Y, given observed ordinal i.i.d. rating samples
x_n drawn from random variable X, and y_n drawn from Y,
X and Y represent two separate Scenarios.
The NAP measure is closely related to the Mann-Whitney statistic for ordinal data.

This module calculates point estimates and approximate confidence intervals for the NAP measure.

There are many ways to estimate a confidence interval for NAP.
Feng et al (2017) studied 29 different methods.
This module uses the "MW-N" variant, defined on page 2607 in their paper,
which showed good performance.
This variant was originally proposed by Newcombe (2009).

*** References:
R. I. Parker and K. Vannest.
An improved effect size for single-case research: Nonoverlap of all pairs.
Behavior Therapy, 40(4):357–367, 2009. doi: 10.1016/j.beth.2008.10.006

D. Feng, G. Cortese, and R. Baumgartner.
A comparison of confidence/credible interval methods for the area under the ROC curve
for continuous diagnostic tests with small sample size.
Statistical Methods in Medical Research, 26(6):2603–2621, 2017.
doi: 10.1177/0962280215602040

R. G. Newcombe.
Confidence intervals for an effect size measure based on the Mann–Whitney statistic.
Part 2: Asymptotic methods and evaluation.
Statistics in Medicine, 25:559–573, 2006. doi: 10.1002/sim.2324

*** Main module function:
nap_statistic --- calculates point estimate and confidence intervals
    for given Attribute-Rating Count arrays

*** Version history:
* Version 0.7:
2021-12-16, first functional version incl. confidence interval
2022-01-03, minor fix to give clearer warning message in case of nan result
"""
import warnings

import numpy as np
from scipy.stats import norm
from scipy.optimize import brentq
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST
# ham.logger.setLevel(logging.DEBUG)  # *** TEST


# -----------------------------------------------------------------
def nap_statistic(x, y, p=0.95):
    """Calculate proportion of Non-overlapping Pairs = NAP result,
    = estimate of P(X > Y), given observed ordinal i.i.d. rating samples
    x_n drawn from random variable X, and y_n drawn from Y,
    where X and Y represent two separate Scenarios.
    :param x: array of outcome COUNTS for samples of ordinal random variable X
        x[i, ...] = number of observed samples of i-th ordinal grade
    :param y: array of COUNTS for ordinal random variable Y, similar
        y.shape == x.shape
    :param p: (optional) scalar confidence level
    :return: nap = array with
        nap[0, ...] = lower confidence-interval limit
        nap[1, ...] = point estimate of P(X > Y)
        nap[2, ...] = upper confidence-interval limit
        nap.shape[1:] == x.shape == y.shape
    """
    result_shape = x.shape[1:]
    cum_y = np.cumsum(y, axis=0)
    # cum_y[i] = n Y <= i
    n_y = cum_y[-1]
    n_x = np.sum(x, axis=0)
    n_x_gt_y = np.sum(x[1:, ...] * cum_y[:-1, ...], axis=0)
    # = sum_i number of pairs where X == i AND Y < i
    n_x_eq_y = np.sum(x * y, axis=0)
    # = n pairs where X == Y
    n_pairs = n_x * n_y
    if np.any(n_pairs == 0):
        logger.warning('NO rating pairs -> undefined NAP value in some case(s).')
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')  # suppress standard warning for div by zero
        nap_point = (n_x_gt_y + 0.5 * n_x_eq_y) / n_pairs
    z_p = - norm.ppf((1. - p) / 2)
    ci_low = np.array([nap_ci_low(p, nx, ny, z_p)
                       for (p, nx, ny) in zip(nap_point.reshape((-1,)),
                                              n_x.reshape((-1,)),
                                              n_y.reshape((-1,)))])
    ci_high = np.array([nap_ci_high(p, nx, ny, z_p)
                       for (p, nx, ny) in zip(nap_point.reshape((-1,)),
                                              n_x.reshape((-1,)),
                                              n_y.reshape((-1,)))])
    return np.array([ci_low.reshape(result_shape),
                     nap_point,
                     ci_high.reshape(result_shape)])


def nap_ci_low(a_hat, nx, ny, z_quantile):
    """Calculate lower limit of symmetric confidence interval for NAP
    :param a_hat: scalar point estimate of P(X > Y)
    :param nx: number of X observations
    :param ny: number of Y observations
    :param z_quantile: quantile at higher tail of standard-normal distribution
    :return: scalar ci_low

    Method: MW-N as described by Feng et al (2017),
        using the improved variant solving their nonlinear Eq. (3)
    """
    def fun_3(a):
        """Function to solve Feng et al Eq (3)
        for the LOWER confidence limit, where 0 <= a < a_hat
        :param a: scalar solution candidate
        :return: scalar function value
        """
        return a_hat - a - z_quantile * nap_stdev(a, nx, ny)
    # --------------------------------------------------
    if nx == 0 or ny == 0:
        return np.nan
    if np.isclose(a_hat, 0.):
        return 0.
    ci_lim, res = brentq(fun_3, 0., a_hat, full_output=True)
    if res.converged:
        return ci_lim
    else:
        logger.warning('Confidence-interval calculation did not converge')
        return np.nan


def nap_ci_high(a_hat, nx, ny, z_quantile):
    """Calculate lower limit of symmetric confidence interval for NAP
    :param a_hat: scalar point estimate of P(X > Y)
    :param nx: number of X observations
    :param ny: number of Y observations
    :param z_quantile: quantile at higher tail of standard-normal distribution
    :return: scalar ci_low

    Method: MW-N as described by Feng et al (2017),
        using the improved variant solving their nonlinear Eq. (3)
    """
    def fun_3(a):
        """Function to solve Feng et al Eq (3)
        for the UPPER confidence limit, where a_hat < a <= 1.
        :param a: scalar solution candidate
        :return: scalar function value
        """
        return a - a_hat - z_quantile * nap_stdev(a, nx, ny)
    # --------------------------------------------------
    if nx == 0 or ny == 0:
        return np.nan
    if np.isclose(a_hat, 1.):
        return 1.
    ci_lim, res = brentq(fun_3, a_hat, 1., full_output=True)
    if res.converged:
        return ci_lim
    else:
        logger.warning('Confidence-interval calculation did not converge')
        return np.nan


def nap_stdev(a, nx, ny):
    """Help function to estimate st.dev. of NAP value
    :param a: scalar candidate value for NAP
    :param nx: number of X observations
    :param ny: number of Y observations
    :return: s = sqrt(var(A)), estimated at point NAP A == a

    Method: var_H defined just before Eq (3) in Feng et al (2017)
    """
    n_star = (nx + ny) / 2 - 1
    # = symmetrized as suggested by Newcombe (2009)
    v = a * (1. - a)
    v *= 1. + n_star * ((1. - a) / (2. - a) + a / (1. + a))
    v /= nx * ny
    return np.sqrt(v)


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    print('*** Testing nap_statistic ***')
    x_count = np.array([1, 2, 3, 4, 5])
    y_count = np.array([3, 3, 3, 3, 3])
    print(f'NAP result =\n\t{nap_statistic(x_count, y_count)}')

    print('*** Testing nap_statistic ***')
    x_count = np.array([0, 0, 0, 5])
    y_count = np.array([4, 3, 2, 0])
    print(f'NAP result =\n\t{nap_statistic(x_count, y_count)}')





"""This module defines a Bayesian probabilistic model of EMA data.
The model is learned using observed EMA recordings from all subjects,
stored in an ema_data.EmaDataSet instance.

This model version uses a mixture of Gaussian distributions
for the parameter vectors in each population of respondents,
represented by one group of subjects,
with different GMMs for each population.

Individual parameter distributions are approximated by sampling.
The population mixture model is common prior for all individuals in
the group recruited from the same population.

The model can NOT use Bayesian sequential learning.
It must be trained in a single run from a single batch of EMA data,
which is usually quite sufficient for practical applications.
However, the learn(...) method can be called several times, if needed.

*** Class Overview:

EmaModel: Defines posterior distributions of scenario probabilities,
    and locations of latent variables for perceptual attribute(s),
    learned from EMA data in an ema_data.EmaDataSet instance.
    The EMA study may include data from different groups of participants,
    represented by a EmaGroupModel instance for each population / participant group.
    Each EmaGroupModel instance contains all individual response-probability models,
    as implemented by an EmaSubjectModel instance for each subject,
    assumed recruited from the SAME population.

ema_group.EmaGroupModel: Defines a GMM for ONE population,
    and contains all individual response-probability models,
    implemented by an EmaSubjectModel instance for each subject,
    in ONE group of test subjects, assumed recruited from the SAME population.

ema_subject.EmaSubjectModel: Distribution of individual parameter vector
    assumed to determine the observed EMA data for ONE subject,
    including, in each EMA record,
    (1) a nominal (possibly multi-dimensional) Scenario category, and
    (2) ordinal Ratings for zero, one, or more subjective Attributes.
    The parameter distribution is represented by an array xi with many samples.

ema_base.EmaParamBase: common properties defining
    indexing into array of parameter vectors, and
    a prior GaussianRV instance for all GMM components.

*** Model Theory: The present model is theoretically similar to
the model of the PairedCompCalc package, partly described in
A. Leijon, M. Dahlquist, and K. Smeds (2019):
Bayesian analysis of paired-comparison sound quality ratings. JASA 146(5):3174–3183.

EXCEPT for the present use of a mixture model for the population(s).

Detailed math documentation is presented in
A Leijon et al. (2022):
Bayesian Analysis of Ecological Momentary Assessment (EMA) Data in python
Journal manuscript in preparation 2022-xx-xx

*** Version History:
* Version 0.8.3:
2022-03-08, minor cleanup

* Version 0.8.1
2022-02-26, cleanup logger output, cleanup comments
2022-02-26, class def EmaGroupModel -> module ema_group, EmaSubjectModel -> ema_subject
2022-02-26, complete separate GMM for each group, GMM components -> EmaGroupModel property comp

* Version 0.8
2022-02-12, Changed VI factorization for better approximation,
    with individual indicators conditional on parameter samples,
    defining variational q(zeta_n, xi_n) = q(zeta_n | xi_n) q(xi_n)

* Version 0.7.2
2022-01-31, Different random Generator for each EmaSubjectModel sampler, spawned from same seed

* Version 0.7.1
2022-01-30, NO module-global random Generator object here
2022-01-30, seed input to EmaModel.initialize, for reproducible results
2022-01-29, random.Generator argument to EmaSubjectModel.initialize
2022-01-11, minor update EmaSubjectModel.adapt_xi, restrict_xi

* Version 0.7:
2021-12-15, new method EmaSubjectModel.mean_attribute_grade
2021-12-20, new method EmaSubjectModel.nap_diff, calculate NAP distance

* Version 0.6:
2021-12-06, Allow user control switches: restrict_attribute OR restrict_threshold
2021-12-08, restrict_attribute: sensory-variable location average forced -> 0.
            restrict_threshold: response-threshold median forced -> 0.

* Version 0.5:
2021-11-03, first functional version
2021-11-04, removed EmaGroupModel.initialize_weights and EmaSubjectModel.initialize_weights
2021-11-05, Hamiltonian sampler -> EmaSubjectModel property -> better stepsize optimization
2021-11-07, store EmaParamBase object as property, so it gets pickled automatically
2021-11-18, cleanup comments
"""
import datetime as dt
import logging

import numpy as np

from EmaCalc.dirichlet_point import JEFFREYS_CONC
# = Jeffreys prior concentration for Dirichlet distribution

from EmaCalc.ema_base import EmaParamBase
from EmaCalc.ema_group import EmaGroupModel
from EmaCalc.ema_latent import Bradley


# -------------------------------------------------------------------
__ModelVersion__ = "2022-02-28"

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# ------------------------------------------------------------------
class EmaModel:
    """Defines probability distributions of EMA recordings
    for all populations from which subjects were recruited,
    and for all subjects in each group,
    given all included EMA data,
    as collected in one ema_data.EmaDataSet instance.

    Model parameters define estimated probabilities for nominal SCENARIOS,
    and estimated latent sensory variables that determine ordinal subjective ratings
    on given ATTRIBUTE questions.
    The sensory variables have continuous values on an INTERVAL scale, as defined by the model,
    although all subjective Attribute ratings are discrete on ORDINAL scales.

    The ATTRIBUTE sensory variables may depend on one or more scenario dimensions,
    as estimated by an ordinal regression model,
    including main and interaction effects as requested by the user.

    The model allows for the possibility that different respondents
    might interpret and use the ordinal rating scales in different ways,
    by including separate response-threshold parameters for each individual.
    """
    def __init__(self, base, groups, rng):
        """
        :param base: single common EmaParamBase object, used by all model parts
        :param groups: dict with EmaGroupModel instances, stored as
            groups[group] = one EmaGroupModel instance, where
                group = dict key = tuple of tuple(s) = (group_factor, factor_category)
            groups[group].subjects[s_id] = an EmaSubjectModel instance, where
            s_id is a key identifying the subject.
        :param rng: random Generator object
        """
        self.base = base
        self.groups = groups
        self.rng = rng

    def __repr__(self):
        return self.__class__.__name__ + '(groups=groups)'

    @classmethod
    def initialize(cls, ds, effects,
                   max_n_comp=None,
                   rv_class=Bradley,
                   restrict_attribute=False,
                   restrict_threshold=False,
                   seed=None):
        """Create a crude initial model from all available count-profile data.
        :param ds: a single ema_data.EmaDataSet instance with all EMA data for analysis
        :param effects: iterable with desired estimated effects of scenario on attribute attribute_grades.
            Each effect element = a key in ds.emf.scenarios, or a tuple of such keys.
        :param max_n_comp: (optional) max number of profile clusters in learned model
            = max number of mixture components in the population model
            The number of actually used components may be reduced during VI learning.
        :param rv_class: (optional) class of latent sensory random variable
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable locations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on response-threshold locations
        :param seed: (optional) integer to get reproducible random sequences
        :return: a cls instance
        """
        ds.ensure_complete()
        if not (restrict_attribute or restrict_threshold):
            logger.warning('Either restrict_attribute or restrict_threshold '
                           + 'should be True, to avoid artificial variance')
        if restrict_attribute and restrict_threshold:
            restrict_attribute = False  # ONLY ONE restriction allowed
            logger.warning(f'Only ONE restriction allowed: using restrict_threshold={restrict_threshold}')
        base = EmaParamBase.initialize(ds.emf, effects, rv_class,
                                       restrict_attribute=restrict_attribute,
                                       restrict_threshold=restrict_threshold)
        # = all base variables, to be used by all model objects
        seed_seq = np.random.SeedSequence(seed)
        rng = np.random.default_rng(seed_seq)
        # = main Generator for all random numbers, EXCEPT EmaSubjectModel instances
        # Same rng for all EmaGroupModel instances, but separate for EmaSubjectModel-s
        group_seeds = seed_seq.spawn(len(ds.groups))
        groups = {g: EmaGroupModel.initialize(max_n_comp, base, g_data, g_seed, rng)
                  for ((g, g_data), g_seed) in zip(ds.groups.items(),
                                                   group_seeds)}
        logger.info('EmaModel initialized with ' +
                    f'{len(groups)} group(s); ' +
                    f'{base.n_parameters} model parameters;\n\t' +
                    f'restrict_attribute = {restrict_attribute}; '
                    f'restrict_threshold = {restrict_threshold};')
        if seed is None:
            logger.debug(f'*** Using seed={seed}')
        else:
            logger.warning(f'*** Using seed={seed} -> reproducible results.')
        return cls(base, groups, rng)

    # ------------------------------------------ General VI learn algorithm:
    def learn(self,
              min_iter=10,
              min_step=0.1,
              max_iter=200,
              max_hours=0.,
              max_minutes=30.,
              callback=None):
        """Learn all individual and population parameter distributions
        from all observed EMA data stored in self.groups[...].subjects[...],
        using Variational Inference (VI).

        This method adapts Gaussian distributions for all mixture components in self.base.comp,
        and Dirichlet distributions for GMM mixture weights in all EmaGroupModel instances,
        and a sampled approximation of parameter distribution in all EmaSubjectModel instances.

        VI maximizes a lower bound to the total likelihood of all observed data.
        The resulting sequence of likelihood values is guaranteed to be non-decreasing,
        except for minor random variations caused by the sampling.

        :param min_iter: (optional) minimum number of learning iterations
        :param min_step: (optional) minimum data log-likelihood improvement,
                 over the latest min_iter iterations,
                 for learning iterations to continue.
        :param max_iter: (optional) maximum number of iterations, regardless of result.
        :param max_hours = (optional) maximal allowed running time, regardless of result.
        :param max_minutes = (optional) maximal allowed running time, regardless of result.
        :param callback: (optional) function to be called after each iteration step.
            If callable, called as callback(self, log_prob)
            where log_prob == scalar last achieved value of VI log-likelihood lower bound
        :return: log_prob = list of log-likelihood values, one for each iteration.

        Result: updated all properties of
        self.groups[...].pop_gmm properties comp, mix_weight
        self.groups[...].subjects[...].xi = parameter sample array
        """
        logger.info(f'Learning the model. Might take some time ...')
        min_iter = np.max([min_iter, 1])
        end_time = dt.datetime.now() + dt.timedelta(hours=max_hours,
                                                    minutes=max_minutes)
        # = last allowed time to start new VI iteration
        log_prob = []
        while (len(log_prob) <= min_iter
               or (log_prob[-1] - log_prob[-1 - min_iter] > min_step
                   and (len(log_prob) < max_iter)
                   and (dt.datetime.now() < end_time))):
            log_prob.append(self.adapt())
            if callable(callback):
                callback(self, log_prob[-1])
            logger.info(f'Done {len(log_prob)} iterations. LL={log_prob[-1]:.2f}')
        if dt.datetime.now() >= end_time:
            logger.warning('Learning stopped at time limit, possibly not yet converged')
        if len(log_prob) >= max_iter:
            logger.warning('Learning stopped at max iterations, possibly not yet converged')
        return log_prob

    def prune(self, min_weight=JEFFREYS_CONC):
        """Prune model to keep only active profile clusters
        :param min_weight: scalar, smallest accepted value for sum individual weight
        :return: None
        Result: all groups pruned with same criterion
        """
        for (g_name, g_model) in self.groups.items():
            # logger.info(f'Pruning group {g_name}:')
            g_model.prune(g_name, min_weight)

    def adapt(self):
        """One VI adaptation step for all groups
        :return: ll = scalar VI lower bound to data log-likelihood,
            incl. negative contributions for parameter KLdiv re priors
        """
        return sum(g_model.adapt(g_name)
                   for (g_name, g_model) in self.groups.items())


# ------------------------------------------------- TEST:
# if __name__ == '__main__':



"""This module includes functions to format output displays of
EmaModel results,
in graphic and textual form.

Some formatting details are defined by module global variables,
which can be modified by user.

*** Version History:
* Version 0.8:
2022-02-15, minor cleanup of tab_percentile: allow cdf=None, header percentile format .1f

* Version 0.7.1:
2022-01-21, minor fix in tab_cred_diff to avoid credibility == 100.0%
2022-01-08, set_tight_layout for all FigureRef objects
2022-01-13, changed EmaDisplaySet.show format argument: show_intervals -> grade_thresholds

* Version 0.7:
2021-12-19, function nap_table to format NAP results

2021-11-07, copied and modified PairedCompCalc -> EmaCalc
"""
# *** set plot y_lim by min-max percentile values, NOT thresholds ???? NO just leave it
# *** NEED format switches for all special chars differing between latex and tab variants ?

import numpy as np
from itertools import cycle, product
import matplotlib.pyplot as plt
import logging

plt.rcParams.update({'figure.max_open_warning': 0})
# suppress warning for many open figures

logger = logging.getLogger(__name__)


# --------------------------- Default format parameters:
FMT = {'colors': 'rbgk',    # to separate results in plots, cyclic
       'markers': 'oxs*_',  # corresponding markers, cyclic use
       'table_format': 'latex',  # or 'tab' for tab-delimited tables
       'figure_format': 'pdf',   # or 'jpg', 'eps', or 'png', for saved plots
       # 'show_intervals': True,  # include median response thresholds in plots
       'grade_thresholds': True,  # include median response thresholds in plots
       'attribute': 'Attribute',  # heading in tables
       'group': 'Group',  # heading in tables
       'subject': 'Subject',  # label for table heads
       'credibility': 'Credibility',  # heading in tables
       }
# = module-global dict with default settings for display details
# that may be changed by user

TABLE_FILE_SUFFIX = {'latex': '.tex', 'tab': '.txt'}
# = mapping table_format -> file suffix


def set_format_param(**kwargs):
    """Set / modify module-global format parameters
    :param kwargs: dict with format variables
        to replace the default values in FMT
    :return: None
    """
    for (k, v) in kwargs.items():
        k = k.lower()
        if k not in FMT:
            logger.warning(f'Format setting {k}={repr(v)} is not known, not used')
        FMT[k] = v


def _percent():
    """Percent sign for tables
    :return: str
    """
    return '\\%' if FMT['table_format'] == 'latex' else '%'


def _gt():
    """> sign for tables
    :return: str
    """
    return '$>$' if FMT['table_format'] == 'latex' else '>'


# ---------------------------- Basic Result Classes
class FigureRef:
    """Reference to a single graph instance
    """
    def __init__(self, ax, path=None, name=None):
        """
        :param ax: Axes instance containing the graph
        :param path: Path to directory where figure has been saved
        :param name: (optional) updated name of figure file
        """
        self.ax = ax
        self.path = path
        self.name = name
        self.ax.figure.set_tight_layout(tight=True)  # ***** utilize all space

    def __repr__(self):
        return (f'FigureRef(ax= {repr(self.ax)}, ' +
                f'path= {repr(self.path)}, name= {repr(self.name)})')

    @property
    def fig(self):
        return self.ax.figure

    def save(self, path, name=None):
        """Save figure to given path
        :param path: Path to directory where figure has been saved
        :param name: (optional) updated name of figure file  **** never used ? ****
        :return: None
        Result: updated properties path, name
        """
        if name is None:
            name = self.name
        path.mkdir(parents=True, exist_ok=True)
        f = (path / name).with_suffix('.' + FMT['figure_format'])
        self.fig.savefig(str(f))
        self.path = path
        self.name = f.name


class TableRef:
    """Reference to a single table instance,
    formatted in LaTeX OR plain tabulated txt versions
    """
    def __init__(self, text=None, path=None, name=None):
        """
        :param text: single string with all table text
        :param path: (optional) Path to directory where tables are saved
        :param name: (optional) updated file name, with or without suffix
            suffix is determined by FMT['table_format'] anyway
        """
        # store table parts instead *****???
        self.text = text
        self.path = path
        self.name = name

    def __repr__(self):
        return (f'TableRef(text= text, ' +    # fmt= {repr(self.fmt)}, ' +
                f'path= {repr(self.path)}, name= {repr(self.name)})')

    def save(self, path, name=None):
        """Save table to file.
        :param path: Path to directory where tables are saved
        :param name: (optional) updated file name, with or without suffix
            suffix is determined by FMT['table_format'] anyway
        :return: None
        Result: updated properties path, name
        """
        if name is None:
            name = self.name
        path.mkdir(parents=True, exist_ok=True)   # just in case
        f = (path / name).with_suffix(TABLE_FILE_SUFFIX[FMT['table_format']])
        if self.text is not None and len(self.text) > 0:
            f.write_text(self.text, encoding='utf-8')
        self.path = path
        self.name = f.name


# ---------------------------------------- Formatting functions:

def fig_percentiles(q_perc,
                    case_labels,  # same input as tab_percentiles
                    # case_order=None,
                    y_label='',
                    file_label='',
                    cat_limits=None,
                    x_offset=0.1,
                    x_space=0.5, **kwargs):
    """create a figure with quality percentile results
    :param q_perc: primary percentile data,
        2D or 3D array with quality percentiles, arranged as
        q_perc[p, c0, c1] = p-th percentile for (case_key_i, category_ci), i= 0, 1,
        OR
        q_perc[p, c0] if only one case dimension is included
    :param y_label: (optional) string for y-axis label
    :param cat_limits: 1D array with response-interval limits (medians)
    :param case_labels: sequence with i-th element = (case_key_i, case_labels_i), where
        case_key_i is key string for i-th case dimension,
        case_labels_i is list of string labels for i-th case dimension
        in the same order as the index order of q_perc, i.e.,
        q_perc[p].size == prod_i len(case_labels_i)
        Thus, q[p, ...,c_i,...] = percentiles for case_labels_i[c_i], i = 0,...,
        **** NOTE: currently only ONE or TWO case dimensions are allowed ***
        Plot will display case_labels_0 along x axis, and case_labels[1:] as sub-variants  *******
    :param file_label: (optional) string as first part of file name
    :param x_offset: (optional) space between case-variants of plots for each x_tick
    :param x_space: (optional) min space outside min and max x_tick values
    :param kwargs: (optional) dict with any additional keyword arguments for plot commands.
    :return: FigureRef instance with plot axis with all results
    """
    # **** input case_labels as dict OR sequence of (key, value) pairs
    # *** generalize for several case dimensions i.e., (c_0,..,c_i,...) ?
    # *** allow user to switch main / sub- dimensions by case_order input ?
    # -------------------------------- check input formats:
    # n_perc = q_perc.shape[0]
    case_shape = tuple(len(c_list) for (c_key, c_list) in case_labels)
    assert q_perc[0].size == np.prod(case_shape, dtype=int), 'mismatching size of q_perc vs. case_list'
    assert q_perc.shape[-1] == case_shape[-1], 'mismatching shape of q_perc vs case_list[-1]'
    if q_perc.ndim == 2:
        q_perc = q_perc[np.newaxis, ...]
        (case_head, case_list) = ('', [''])
    else:
        q_perc = q_perc.transpose((2, 0, 1))
        (case_head, case_list) = case_labels[1]
    # q_perc now indexed as q_perc[case, p, object], prepared to plot one case at a time
    (x_label, x_tick_labels) = case_labels[0]
    # ------------------------------------------------------------------
    fig, ax = plt.subplots()
    x = np.arange(0., len(x_tick_labels)) - x_offset * (len(case_list) - 1) / 2
    for (y, c_label, c, m) in zip(q_perc, case_list,
                                  cycle(FMT['colors']), cycle(FMT['markers'])):
        label = case_head + '=' + c_label
        if y.shape[0] == 1:
            ax.plot(x, y[0, :],  # marker at single point percentile,
                    linestyle='',
                    marker=m, markeredgecolor=c, markerfacecolor='w',
                    label=label,
                    **kwargs)
        elif y.shape[0] == 2:  # plot min,max line, with marker at ends
            line = ax.plot(np.tile(x, (2, 1)),
                           y,
                           linestyle='solid', color=c,
                           marker=m, markeredgecolor=c, markerfacecolor='w',
                           **kwargs)
            line[0].set_label(label)
        else:
            ax.plot(np.tile(x, (2, 1)),
                    y[[0, -1], :],  # plot min, max line, no markers
                    linestyle='solid', color=c,
                    **kwargs)
            line = ax.plot(np.tile(x, (y.shape[0] - 2, 1)),
                           y[1:-1, :],
                           linestyle='solid', color=c,  # markers only at inner percentiles
                           marker=m, markeredgecolor=c, markerfacecolor='w',
                           **kwargs)
            line[0].set_label(label)
            # set only ONE label, even if several points
        x += x_offset
    (x_min, x_max) = ax.get_xlim()
    x_min = min(x_min, -x_space)
    x_max = max(x_max, len(x_tick_labels) - 1 + x_space)
    ax.set_xlim(x_min, x_max)
    if cat_limits is not None and FMT['grade_thresholds']:
        _plot_response_intervals(ax, cat_limits)
    ax.set_xticks(np.arange(len(x_tick_labels)))
    ax.set_xticklabels(x_tick_labels)
    ax.set_ylabel(y_label)  # plain without unit
    ax.set_xlabel(x_label)
    if np.any([len(cl) > 0 for cl in case_list]):
        ax.legend(loc='best')
    if len(file_label) > 0:
        file_label += '_'
    f_name = file_label + '*'.join(c_key for (c_key, c_val) in case_labels)
    return FigureRef(ax, name=f_name)


def _plot_response_intervals(ax, c_lim):
    """plot horizontal lines to indicate response-category intervals
    :param ax: axis object
    :param c_lim: 1D array with scalar interval limits
    :return: None
    """
    (x_min, x_max) = ax.get_xlim()
    return ax.hlines(c_lim, x_min, x_max,
                     linestyle='solid',
                     colors='k',
                     linewidth=0.2)


# ----------------------------------------- table displays:

def tab_percentiles(q_perc,
                    perc,
                    case_labels,
                    case_order=None,
                    cdf=None,
                    y_label='',  # *** replaced by file_label ***
                    file_label=''):
    """Create table with quality percentile results.
    This function is general and can handle any dimensionality of the data.
    :param q_perc: min 2D array with quality percentiles, stored as
        q_perc[p, c0,...,ci,...] = p-th percentile in (c0,...,ci,...)-th case condition
        OR any other array indexing with same size and same element order
        q_perc.shape[0] == len(perc)
        q_perc.size == len(perc) * n_rows; with n_rows as defined by case_list
    :param perc: list of percentage values in range 0-100
        len(perc) == q_perc.shape[0]
    :param y_label: string with label of tabulated percentiles, NOT USED
    :param case_labels: sequence OR dict with elements (case_factor_i, case_labels_i), where
        case_factor_i is key string for i-th case dimension,
        case_labels_i is list of string labels for i-th case dimension
        in the same order as the index order of q and cdf, i.e.,
        len(case_labels_i) == q.shape[i+1] == cdf.shape[i], if full multi-dim index is used
        Thus, q[p, ...,c_i,...] = percentiles for case_labels_i[c_i], i = 0,...,
    :param case_order: (optional) sequence of case_label keys, one for each case column
        len(case_order) == len(case_labels)
        Table columns are ordered as defined by case_order, if specified,
        otherwise columns are ordered as case_labels.keys()
    :param cdf: (optional) min 1D array with cumulative distribution values at zero,
        cdf[c0,...,ci,...] = probability that y-variable <= 0 in (c0,...,ci,...)-th case
        OR any other array indexing with same size and same element order
        cdf.size == number of rows as defined by case_list
    :param file_label: (optional) string as first part of file name
    :return: TableRef object, with header + one line for each combination of case labels.
        Number of table rows == prod q.shape[:-1] == prod_i len(case_labels_i)
    """
    def make_row(c, q, p_0=None):
        """make cells for one table row
        :param c: case tuple
        :param q: percentile quality values
        :param p_0: (optional) scalar probability q <= 0
        :return: row = list with cells
            len(row) == 1 + len(c) + len(q) + 2
        """
        c_dict = dict(c)
        # already in right order, need only c.values() ******
        return ([c_dict[c_head] for c_head in case_order]
                + [f'{p:.2f}' for p in q]
                + ([] if p_0 is None
                   else [f'{p_0*100:.0f}' + _percent(),
                         f'{(1.-p_0)*100:.0f}' + _percent()])
                )

    # --------------------------------------------------------------------
    case_labels = dict(case_labels)  # just in case it was a list
    if case_order is None:
        case_order = [*case_labels.keys()]
    assert len(case_order) == len(case_labels), 'Incompatible len(case_order) != len(case_list)'
    assert all(c in case_labels for c in case_order), 'Some case_order not in case_list'
    case_shape = tuple(len(c_labels) for c_labels in case_labels.values())
    n_rows = np.prod(case_shape, dtype=int)
    # = number of table rows as defined by case_list
    assert len(perc) == q_perc.shape[0], 'Incompatible q_perc.shape[0] != n of percentiles'
    assert n_rows == np.prod(q_perc.shape[1:], dtype=int), 'Incompatible size of case_list and q_perc'
    if cdf is not None:
        assert n_rows == cdf.size, 'Incompatible size of case_list and cdf'
    # -------------------------------------------------------------------
    # transpose q_perc, cdf to case_order index order:
    q_perc = np.moveaxis(q_perc, 0, -1)
    q_perc = q_perc.reshape((*case_shape, len(perc)))
    # q_perc[c_0, ..., c_i, ..., p] = p-th percentile in (c0,...,ci,...)-th case
    case_keys = [*case_labels.keys()]
    case_axes = tuple(case_keys.index(c) for c in case_order)
    q_perc = q_perc.transpose((*case_axes, -1))
    q_perc = q_perc.reshape((n_rows, len(perc)))
    if cdf is not None:
        cdf = cdf.reshape(case_shape)
        # cdf[c_0, ..., c_i, ...] = prob(y-variable <=0) in (c0,...,ci,...)-th case
        cdf = cdf.transpose(case_axes)
        cdf = cdf.reshape((-1,))
    # --------------------------------------------------------------------
    align = [len(case_order) * 'l', len(perc) * 'r', 2 * 'r']
    h = [*case_order] + [f'{p:.1f}' + _percent() for p in perc]
    if cdf is not None:
        h += ['<= 0', '> 0']
    case_rows = product(*(product([c_i], case_labels[c_i])
                          for c_i in case_order))
    if cdf is None:
        rows = [make_row(c, p)
                for (c, p) in zip(case_rows,
                                         q_perc)
                ]
    else:
        rows = [make_row(c, p, cdf_0)
            for (c, p, cdf_0) in zip(case_rows,
                                     q_perc,
                                     cdf)
            ]
    if len(file_label) > 0:
        file_label += '_'
    f_name = file_label + '*'.join(case_order)
    return TableRef(_make_table(h, rows, align), name=f_name)


def tab_credible_diff(diff,
                      diff_labels,
                      diff_head=tuple(),  # *** not used OR define diff_column order *****
                      case_labels=tuple(),
                      case_head=tuple(),   # *** not used OR define diff_column order *****
                      y_label='',
                      file_label=''):
    """Create table with credible differences among results
    :param diff: list of tuples ((i,j), p) OR ((i,j,c), p),
        defining jointly credible differences, indicating that
        prob{ quality of diff_labels[i] > quality of diff_labels[j]}, OR
        prob{ quality of diff_labels[i] > quality of diff_labels[j] | case_labels[c] }
        AND all previous pairs } == p
    :param diff_labels: iterator or sequence of dicts or tuples of (key, value) pairs
        with labels of compared random-vector elements, yielding
        diff_labels[i] = ((key0, value0), (key1, value1),...), defining
        diff[...] == ((i,j, c), p) <=> diff_labels[i] > diff_labels[j] with prob p
        All diff_labels elements MUST have the SAME ORDER of the keys
        len(diff_labels) == max possible diff category index (i, j)
    :param case_labels: (optional) iterator or sequence of dicts or tuples of (key, value) pairs,
        yielding
        case_labels[c] == (case_label1, case_label2, ...), such that
        diff[...] == ((i,j, c), p) <=> diff[...] is valid given case_labels[c]
        len(case_labels) == max possible case index c in diff
        len(case_list[j]) == len(case_head), for all j
    :param diff_head: (optional) sequence with all diff_labels key(s) for heading column(s) with diff pairs
    :param case_head: (optional) sequence of case factors, one for each case-dimension table column
        len(case_heads) == len(c) for each c in case_labels
    :param y_label: (optional) string with label of tabulated attribute
    :param file_label: (optional) string for first part of file name
    :return: TableRef object with header lines + one line for each credible difference,
    """
    # **** use diff_head and case_head input to change column order ? **************
    if len(diff) == 0:
        return None
    diff_labels = [*(dict(diff_label_i)  # ******** or require dict input ? ***********
                     for diff_label_i in diff_labels)]
    case_labels = [*(dict(case_label_i)  # ******** or require dict input ? ***********
                     for case_label_i in case_labels)]
    if len(diff_head) == 0:
        diff_head = [*diff_labels[0].keys()]  # ****** not used
    if len(case_labels) > 0 and len(case_head) == 0:
        case_head = [*case_labels[0].keys()]  # ****** not needed ?
    align = ['l', len(diff_head) * 'l', 'c', len(diff_head) * 'l', len(case_head) * 'l', 'r']
    y = ''
    if len(y_label) > 0:
        y = y_label + ':'
    head = [y, *diff_head,
            _gt(), *diff_head,
            *case_head,
            FMT['credibility'] + ' (' + _percent() + ')']
    rows = []
    for (d, p) in diff:
        if len(d) == 2:
            c_dict = dict()
        else:
            c_dict = case_labels[d[2]]
        if p < 0.9995:
            s_p = f'{100*p:.1f}'
        else:
            s_p = '>99.9'  # avoid rounding to 100.0
        rows.append(['AND', *diff_labels[d[0]].values(),
                     _gt(), *diff_labels[d[1]].values(),
                     *c_dict.values(),
                     s_p])
    rows[0][0] = ''  # No AND on first row
    if len(file_label) > 0:
        file_label += '_'
    f_name = file_label + '*'.join(diff_head) + '-diff'
    if len(case_head) > 0:
        f_name += '_' + '*'.join(case_head)
    return TableRef(_make_table(head, rows, align), name=f_name)


def nap_table(nap, attribute, sc_diff, case_labels, conf_level):
    """Make table of NAP results, with point estimate and confidence interval
    :param nap: dict with elements (s, s_nap), where
        s = subject id key,
        s_nap = array with NAP results in [0, 1], stored as
        s_nap[0, ...] = lower confidence interval limit(s)
        s_nap[1, ...] = point estimate(s) of NAP
        s_nap[2, ...] = upper confidence interval limit(s)
        s_nap.shape == (3, len(c_i) for c_i in case_labels.values())
    :param attribute: attribute label
    :param sc_diff: single scenario dimension key for which s_nap is calculated
    :param case_labels: dict with (scenario_key, scenario_cat_list)
        corresponding to s_nap.shape
    :param conf_level: scalar value
    :return: a TableRef instance with all results
    """
    def subject_rows(s, s_nap):
        """format all rows for ONE subject, one for each case-label combination
        :param s: subject id
        :param s_nap: array with (low, mid, high) NAP estimates
            s_nap[..., i] = result for i-th case
        :return: list of row lists
        """
        s_rows = [[s] + list(case) + [f'{nap_i:.2f}'
                                      for nap_i in lmh]
                  for (lmh, case) in zip(s_nap.reshape((3, -1)).T,
                                         product(*case_labels.values()))]
        for r_i in s_rows[1:]:
            r_i[0] = ' '  # suppress subject label except on first line
        return s_rows
    # -------------------------------------------

    align = 'l' + len(case_labels) * 'l' + 3 * 'r'
    c_low = (1. - conf_level) / 2
    c_high = 1. - c_low
    head = [FMT['subject'], *case_labels.keys(), f'{c_low:.1%}', 'NAP', f'{c_high:.1%}']
    rows = []
    for (s, s_nap) in nap.items():
        rows.extend(subject_rows(s, s_nap))
    f_name = '_'.join([attribute, sc_diff])
    if len(case_labels) > 0:
        f_name += '_' + '*'.join(case_labels.keys())
    return TableRef(_make_table(head, rows, align), name=f_name)  # *********** temp


# ------------------------------------------ internal help functions:
# more variants may be added for other table formats

table_begin = {'latex': lambda align: '\\begin{tabular}{' + ' '.join(c for c in align) + '}\n',
               'tab': lambda align: ''}
table_head_sep = {'latex': '\\hline\n',
                  'tab': ''}
table_cell_sep = {'latex': ' & ',
                  'tab': ' \t '}
table_row_sep = {'latex': '\\\\ \n',
                 'tab': '\n'}
table_end = {'latex': '\\hline\n\\end{tabular}',
             'tab': ''}


def _make_cell(text, col_span, fmt):
    """Format multi-column table cell, usually only for header line
    :param text: cell contents
    :param col_span: number of columns to span
    :param fmt: str key for table format
    :return: string with latex or plain cell contents
    """
    if fmt == 'latex' and col_span > 1:
        return '\\multicolumn{' + f'{col_span}' + '}{c}' + '{' + text + '}'
    else:
        return text


def _make_table(header, rows, col_alignment):
    """Generate a string with table text.
    :param header: list with one string for each table column
    :param rows: list of rows, ehere
        each row is a list of string objects for each column in this row
    :param col_alignment: list of alignment symbols, l, r, or c
        len(col_alignment) == len(header) == len(row), for every row in rows
    :return: single string with complete table
    """
    def make_row(cells, fmt):
        return table_cell_sep[fmt].join(f'{c}' for c in cells) + table_row_sep[fmt]
    # ------------------------------------------------------------------------

    fmt = FMT['table_format']  # module global constant
    t = table_begin[fmt](col_alignment)
    t += table_head_sep[fmt]
    t += make_row(header, fmt)
    t += table_head_sep[fmt]
    t += ''.join((make_row(r, fmt) for r in rows))
    t += table_end[fmt]
    return t



"""This module defines a class for individual subject parameter model
to be part of an ema_group.EmaGroupModel instance,
which is part the main Bayesian probabilistic model of EMA data.

Individual parameter distributions are approximated by sampling.
The population mixture model is common prior for all individuals in
the group recruited from the same population.

*** Class Defined here:

EmaSubjectModel: Distribution of individual parameter vector
    assumed to determine the observed EMA data for ONE subject,
    including, in each EMA record,
    (1) a nominal (possibly multi-dimensional) Scenario category, and
    (2) ordinal Ratings for zero, one, or more subjective Attributes.
    The parameter distribution is represented by an array xi with many samples.

*** Version History:
* Version 0.8.3:
2022-03-08, minor cleanup logging to work in multi-processing

* Version 0.8.2: prepared for multi-processing subject adapt() in parallel processes
2022-03-03, EmaSubjectModel methods mean_zeta, mean_zeta_mom no longer needed

* Version 0.8.1: minor cleanup of comments and logger output

* Version 0.8
2022-02-12, Changed VI factorization for better approximation,
    with individual indicators conditional on parameter samples,
    defining variational q(zeta_n, xi_n) = q(zeta_n | xi_n) q(xi_n)
"""
# *** EmaSubjectModel use BoundedHamiltonianSampler to avoid numerical overflow ? ***
import multiprocessing, logging
import warnings

import numpy as np
from scipy.optimize import minimize
from scipy.special import logit, logsumexp, softmax

from samppy import hamiltonian_sampler as ham
from samppy.sample_entropy import entropy_nn_approx as entropy

from EmaCalc.dirichlet_point import JEFFREYS_CONC
# = Jeffreys prior concentration for Dirichlet distribution

from EmaCalc.ema_base import PRIOR_PARAM_SCALE
from EmaCalc.ema_base import response_thresholds, d_response_thresholds
from EmaCalc.ema_base import tau_inv
from EmaCalc.ema_nap import nap_statistic


# -------------------------------------------------------------------
__ModelVersion__ = "2022-03-08"

DITHER_PARAM_SCALE = 0.1 * PRIOR_PARAM_SCALE
# -> initial dithering of point-estimated individual parameters

N_SAMPLES = 1000
# = number of parameter vector samples in each EmaSubjectModel instance

logger = logging.getLogger(__name__)
# logger does NOT inherit parent handlers, when this module is running as child process
if multiprocessing.current_process().name != "MainProcess":
    # restore a formatter like ema_logging, only for console output
    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter('{asctime} {name}: {message}',
                                           style='{',
                                           datefmt='%H:%M:%S'))
    logger.addHandler(console)
# Might instead send the parent logger as argument to adapt() ? ***
# logger.setLevel(logging.DEBUG)  # *** TEST
# ham.logger.setLevel(logging.DEBUG)  # *** TEST sampler


# -------------------------------------------------------------------
class EmaSubjectModel:
    """Container for EMA response data for ONE respondent,
    and a sampled approximation of the individual parameter distribution,
    and corresponding mixture weights for the parameters.

    Individual parameter distributions are approximated by a large set of samples
    stored as property xi, with
    self.xi[s, :] = s-th sample vector of parameters,
    with subsets of parameter types (alpha, beta, eta) as defined by self.base.
    """
    # *** subject_id as property of EmaSubjectModel ? ********
    def __init__(self, base, scenario_count, rating_count, xi, rng, prior=None):
        """
        :param id: key to identify self, instead of group.dict key ???
        :param base: single common EmaParamBase object, used by all model parts
        :param scenario_count: 2D array with response counts
            scenario_count[k0, k] = number of responses
            in k-th <=> (k1, k2, ...)-th scenario category at k0-th test stage,
            using flattened index for scenario dimensions 1,2,....
            NOTE: ema_data.EmaFrame always stores test stage as first scenario dimension.
        :param rating_count: list of 2D arrays with response counts
            rating_count[i][l, k] = number of responses for i-th ATTRIBUTE,
            at l-th ordinal level, given the k-th <=> (k0, k1, k2, ...)-th scenario
        :param xi: 2D array with parameter sample vector(s)
            xi[s, j] = s-th sample of j-th individual parameter,
                concatenating parameter sub-types as defined in superclass.
        :param rng: random Generator object for sampler
        :param prior: ref to owner EmaGroupModel  *** -> arg for self.adapt() ???
        """
        # self.id = id
        self.base = base
        self.scenario_count = scenario_count
        self.rating_count = rating_count
        self.xi = xi
        self._sampler = ham.HamiltonianSampler(x=self.xi,
                                               fun=self._neg_ll,
                                               jac=self._grad_neg_ll,
                                               epsilon=0.2,
                                               n_leapfrog_steps=10,     # = default
                                               min_accept_rate=0.8,     # = default
                                               max_accept_rate=0.95,    # = default
                                               rng=rng
                                               )
        # keeping sampler properties across learning iterations
        self.prior = prior
        self.ll = None  # space for log-likelihood result from self.adapt()

    def __repr__(self):
        return (self.__class__.__name__ + '('
                + '\n\tscenario_count=' + f'{self.scenario_count},'
                + '\n\trating_count=' + f'{self.rating_count},'
                + f'\n\txi= parameter array with shape {self.xi.shape}; id={id(self.xi)},'
                + f'\n\tid(prior)= {id(self.prior)},'
                + f'\n\tll= {self.ll})')

    @classmethod
    def initialize(cls, base, ema_data, rng):  # *** rng -> seed_seq ?
        """Create model from recorded data
        :param base: single common EmaParamBase object
        :param ema_data: list of ema tuples, as stored im ema_data.EmaDataSet instance
        :param rng np.random.Generator for sampler use
        :return: a cls instance
        """
        z = _count_scenarios(ema_data, base.emf)
        # z[t, k] = number of EMA assessments at t-th test stage
        # in k-th <=> (k1, k2, ...)-th scenario, EXCL. k0= test stage
        y = _count_ratings(ema_data, base.emf)
        # y[i][l, k] = number of attribute_grades at l-th ordinal level for i-th ATTRIBUTE question
        # given k-th <=> (k0, k1, k2, ...)-th scenario (INCL. k0= test stage)
        alpha = _initialize_scenario_param(z)
        xi = list(np.ravel(alpha))
        for y_i in y:
            theta = _initialize_rating_theta(y_i).reshape(-1)
            a = base.theta_map.reshape((base.theta_map.shape[0], -1))  # **** needed ?
            beta = np.linalg.lstsq(a.T, theta, rcond=None)[0]
            eta = np.zeros(len(y_i))
            xi.extend(np.concatenate((beta, eta)))
        xi = np.array(xi)
        # dither to N_SAMPLES:
        xi = xi + DITHER_PARAM_SCALE * rng.standard_normal(size=(N_SAMPLES, len(xi)))
        # w = np.ones((len(xi), len(base.comp))) / len(base.comp)  # = None? ****** uniform distribution
        return cls(base, z, y, xi, rng)

    def _kl_div_zeta(self, prior):
        """KLdiv for indicator variables zeta re prior
        = KLdiv{q(zeta | self.xi) || p(zeta | prior.mix_weight)}
        using current variational q(xi, zeta) = q(zeta | xi) q(xi)
        :param prior: ref to EmaGroupModel instance containing self
        :return: kl = scalar E{ ln q(zeta | self.xi) - ln p(zeta | prior.mix_weight)}

        Method: Leijon doc eq:VIlowerBoundCalc
        """
        w = prior.mean_conditional_zeta(self.xi)
        # w[c, s] = E{zeta_c | self.xi[s, :]} Leijon doc eq:ProbZetaGivenXi
        return np.sum(np.mean(w * (np.log(w + np.finfo('float').tiny)  # avoid log(0.) = nan
                                   - prior.mix_weight.mean_log[:, None]),
                              axis=1))

    def adapt(self, s_name):
        """Adapt parameter distribution self.xi
        to stored EMA count data, given the current self.w
        and the current estimate of population GMM components self.base.comp.
        :param s_name: subject id, for logger output
        :return: self, to send result via map() or Pool.imap()

        Result: Updated self.xi,
            self.ll = E{ ln p(self.scenario_count, self.rating_count | self.xi) }_q(xi)
             - E{ ln q(xi) / prior_p(xi) }_q(xi)
             - KLdiv( q(zeta | xi) || p(zeta || prior.mix_weight)
        """
        # find MAP point first:
        xi_0 = np.mean(self.xi, axis=0)
        res = minimize(fun=self._neg_ll,  # _1,
                       jac=self._grad_neg_ll,  # _1,
                       x0=xi_0)
        if res.success:
            xi_map = res.x.reshape((1, -1))
        else:
            raise RuntimeError(f'{s_name}: MAP search did not converge: '
                               + 'res= ' + repr(res))
        if len(self.xi) != N_SAMPLES:
            # run sampler starting from x_map
            self._sampler.x = xi_map
        else:
            # we have sampled before, start from those samples
            self._sampler.x = self.xi + xi_map - xi_0
        self._sampler.safe_sample(n_samples=N_SAMPLES, min_steps=2)
        self.xi = self._sampler.x
        self.restrict_xi()
        self._sampler.U = self._sampler.potential(self._sampler.x)
        self._sampler.args = ()  # just in case a pickled prior copy was there
        lp_xi = - np.mean(self._sampler.U)
        # lp_xi = E_xi{ ln p(data | xi) + self.prior.logpdf(xi) }
        h_xi = entropy(self.xi)
        # approx = - E{ ln q(xi) }
        kl_zeta = self._kl_div_zeta(self.prior)
        self.ll = lp_xi + h_xi - kl_zeta
        logger.debug(f'{s_name}: adapt: ll={self.ll:.3f}; '
                     + f'(lp_xi={lp_xi:.3f}; '
                     + f'h_xi= {h_xi:.3f}. '
                     + f'-kl_zeta= {-kl_zeta:.3f})')
        self.ll = lp_xi + h_xi - kl_zeta
        return self

    def restrict_xi(self):
        """Force mean(theta) -> 0, OR mean(tau) -> 0, after sampling
        for all attributes.
        :return: None
        Result: self.xi = self._sampler.x modified in place
            theta and tau samples modified by same amount
        """
        n_samples = self.xi.shape[0]
        # *** restrict log scenario prob -> normalized prob:
        for sc_slice in self.base.scenario_slices:
            d = logsumexp(self.xi[:, sc_slice], axis=-1, keepdims=True)
            # logger.debug(f'xi scenario restrict d: mean={np.mean(d):.3}; std={np.std(d):.3}; '
            #              + f'max={np.amax(d):.3}; min={np.amin(d):.3}')
            self.xi[:, sc_slice] -= d
        for (a, a_slice) in self.base.attribute_slice_dict.items():
            theta = self.base.attribute_theta(self.xi, a)
            tau = self.base.attribute_tau(self.xi, a)
            if self.base.restrict_attribute:
                d = np.mean(theta.reshape((n_samples, -1)),
                            axis=-1, keepdims=True)
            elif self.base.restrict_threshold:
                d = np.median(tau, axis=-1, keepdims=True)
            else:
                d = 0.
            # d = random offset to be zero-ed out
            n_beta_0 = self.base.beta_0_size()
            # number of beta parameters = first part of a_slice
            beta_slice = slice(a_slice.start,
                               a_slice.start + n_beta_0)
            # = beta slice for FIRST regression effect term
            # logger.debug(f'xi attribute restrict d: mean={np.mean(d):.3}; std={np.std(d):.3}; '
            #              + f'max={np.amax(d):.3}; min{np.amin(d):.3}')
            self.xi[:, beta_slice] -= d
            # adjust tau
            eta = tau_inv(tau - d)
            eta_slice = slice(a_slice.start + self.base.n_beta, a_slice.stop)
            # **** method self.base.eta_slice ???
            self.xi[:, eta_slice] = eta

    def _neg_ll(self, xi):
        """Objective function for self.adapt_xi
        :param xi: 1D or 2D array of candidate parameter vectors
        :return: neg_ll = scalar or 1D array
            neg_ll[...] = - ln P{ self.scenario_count, self.rating_count | xi[..., :]}
                    - ln p(xi[..., :] | prior)
            neg_ll.shape == xi.shape[:-1]
        """
        return - self.prior.logpdf(xi) - self.logprob(xi)

    def _grad_neg_ll(self, xi):
        """Gradient of self._neg_ll w.r.t. xi
        :param xi: 1D or 2D array of candidate parameter vectors
        :return: dll = scalar or 1D array
            dll[..., j] = d _neg_ll(xi[..., :]) / d xi[..., j]
            dll.shape == xi.shape
        """
        return - self.prior.d_logpdf(xi) - self.d_logprob(xi)

    def logprob(self, xi):
        """log likelihood of EMA count data, given parameters xi
        :param xi: 1D or 2D array of candidate parameter vectors
        :return: ll = scalar or 1D array
            ll[...] = ln P{ self.scenario_count, self.rating_count | xi[..., :]}
            ll.shape == xi.shape[:-1]
        """
        def scenario_logprob(sc_xi):
            """Log probability mass for scenario categories in ONE test stage
            :param sc_xi: 1D or 2D array with model parameters for ONE test stage
                = alpha = log P{scenarios}, NOT necessarily normalized
            :return: lp = 1D or 2D array with
                lp[..., k] = log P{ k-th scenario | sc_xi[..., :] }
                properly normalized, with sum_k exp(lp[:, k]) == 1.
                lp.shape == sc_xi.shape
            """
            # **** consider using log-Dirichlet here ???
            # NO, this would cause an extra layer in the hierarchy for scenarios,
            # with a log-normal distribution of concentration parameters
            # while these only indirectly determine the probabilities.
            # In the present model, the alpha params directly determine the probabilities.
            return sc_xi - logsumexp(sc_xi, axis=-1, keepdims=True)

        def rating_logprob(a_xi):
            """Log probability mass for rating categories for ONE attribute.
            :param a_xi: 1D or 2D array with model parameters for this attribute
            :return: lp = 2D or 3D array with
                lp[..., l, k] = log P{ l-th ordinal level | k-th scenario, a_xi[..., :] }
            """
            return self.base.rv_class.log_cdf_diff(*self.cdf_args(a_xi))
        # ---------------------------------------------------------------

        ll_z = sum(np.dot(scenario_logprob(xi[..., sc_t]), z_t)
                   for (z_t, sc_t) in zip(self.scenario_count,
                                          self.base.scenario_slices))
        # z_t[k] = k-th scenario count for t-th test stage
        # sc_t = parameter slice for t-th test stage
        ll_y = sum(np.einsum('lk, ...lk -> ...',  # ********* use np.tensordot ???
                             y_i, rating_logprob(xi[..., a_i]))
                   for (y_i, a_i) in zip(self.rating_count,
                                         self.base.attribute_slices))
        # y_i[l, k] = (l, k)-th count for i-th attribute rating
        # a_i = param slice for i-th attribute rating
        return ll_z + ll_y

    def d_logprob(self, xi):
        """Jacobian of self.logprob(xi)
        :param xi: 1D or 2D array with candidate parameter vectors
        :return: d_ll = 2D array
            d_ll[..., j] = d ln P{ self.scenario_count, self.attribute_grades | xi[..., :]} / d xi[..., j]
        """
        def d_scenario_logprob(sc_xi):  # ******** use count input, save RAM ?
            """Gradient of scenario_logprob for ONE test stage
            :param sc_xi: 1D or 2D array with model parameters for ONE test stage
                = alpha = log P{scenarios}, NOT necessarily normalized
            :return: dlp = 2D or 3D array with
                dlp[..., k, j]
                = d log P{ scenario_logprob()[..., k] | sc_xi[..., :] } / d sc_xi[..., j]
            """
            # lp[s, k] = sc_xi[s, k] - ln sum_i(exp(sc_xi[s, i]))
            # dlp[s, k, j] = 1(k==j) - softmax(sc_xi[s, :], axis=-1)[s, j]
            return np.eye(sc_xi.shape[-1]) - softmax(sc_xi[..., None, :], axis=-1)

        # -----------------------------------------------------------------------

        d_ll = [np.einsum('k, ...kj -> ...j',  # ******* use dot or matmul?
                          z_t, d_scenario_logprob(xi[..., sc_t]))
                for (z_t, sc_t) in zip(self.scenario_count,
                                       self.base.scenario_slices)]
        # z_t[k] = k-th count for t-th test stage
        # sc_t = parameter slice for t-th test stage
        for (y_i, a_i) in zip(self.rating_count, self.base.attribute_slices):
            # a_xi = xi[:, a_i]
            a_xi = xi[..., a_i]
            (dlp_low, dlp_high) = self.base.rv_class.d_log_cdf_diff(*self.cdf_args(a_xi))
            # append dlp_d_beta[s, j] = d rating_logprob()[s] / d xi[:, a_i][:, j]
            d_ll.append(np.einsum('...lk, lk, jk -> ...j',
                                  dlp_low + dlp_high, y_i, - self.base.theta_map))  # tensordot 2 steps?
            d_tau_d_eta = d_response_thresholds(a_xi[..., self.base.n_beta:])
            # append dlp_d_eta[..., j] = d rating_logprob()[s] / d xi[:, a_i][:, self.n_beta + j]
            d_ll.append(np.einsum('...lk, lk, ...lj -> ...j',  # *** use dot , matmul ?
                                  dlp_low, y_i, d_tau_d_eta[..., :-1, :]) +
                        np.einsum('...lk, lk, ...lj -> ...j',
                                  dlp_high, y_i, d_tau_d_eta[..., 1:, :]))
            # all d_ll[i][..., :] elements have matching shapes
        return np.concatenate(d_ll, axis=-1)

    def cdf_args(self, a_xi):
        """Extract arguments for rv_class logprob calculation
        :param a_xi: 1D or 2D array with rating-model sample for given attribute a
            a_xi[..., j] = ...-th sample of j-th rating-model parameter
        :return: tuple (arg_low, arg_high) with 2D or 3D array args for probabilistic model,
            such that
            P[ l-th response | ...-th parameter sample, k-th scenario ] =
            = rv_class.cdf(arg_high[..., l, k]) - rv_class.cdf(arg_low[..., l, k])
        """
        theta = np.dot(a_xi[..., :self.base.n_beta], self.base.theta_map)
        # theta[..., k] = ...-th sample of latent variable, given k-th scenario
        # tau_old = self.base.tau(a_xi[..., self.base.n_beta:])
        tau = response_thresholds(a_xi[..., self.base.n_beta:])
        a = tau[..., :-1, None]  # lower interval limits
        b = tau[..., 1:, None]  # upper interval limits
        return a - theta[..., None, :], b - theta[..., None, :]

    def rvs(self, size=N_SAMPLES):
        # re-sample if size != len(self.xi) ???
        return self.xi

    def mean_attribute_grade(self, sc=None):
        """Average raw attribute grades
        :param sc: scenario key, or tuple of scenario keys for result
            grade counts summed across scenario dimensions NOT included in sc
        :return: dict with elements (a, theta), where
            a = attribute key, from self.emf.attribute_grades
            th = mD array with
            th[i0, i1, ...] = mean grade, given scenario
            (self.scenarios[sc[0]][i0], self.scenarios[sc[1]][i1], ...)
            aggregated across other scenario dimensions
        """
        # *** call this only for ONE given attribute at a time, like nap_diff ??? *****
        def mean_grade(a_count, count_axes):
            """Average grade for ONE attribute
            :param a_count: mD array with
                a_count[l, k0, k1, ...] = count of l-th grade, given (k0, k1, ...)-th scenario
            :param count_axes: tuple with a_count axes to keep in result
            :return: g = array with mean grades
                g[i0, i1, ...] = mean grade, given (i0, i1, ...)-th category
                    in self.attribute_rating count_axes, aggregated across other scenario axes
                g.shape == self.base.emf.scenario_shape
            """
            a_count = np.moveaxis(a_count, count_axes,
                                  tuple(range(1, 1 + len(count_axes))))
            sum_axis = tuple(range(1 + len(count_axes), len(a_count.shape)))
            a_count = np.sum(a_count, axis=sum_axis)  # across all other axes
            grades = [1 + i for i in range(a_count.shape[0])]
            g = np.tensordot(grades, a_count,
                             axes=1)
            n_count = np.sum(a_count, axis=0)
            if np.any(n_count == 0):
                logger.warning('NO ratings -> undefined mean in some case(s)')
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')  # suppress standard warning for div by zero
                return g / n_count
        # ----------------------------------------------------------------
        sc_keys = list(self.base.emf.scenarios.keys())
        if sc is None:
            sc = tuple(sc_keys)  # include all scenario dimensions
        if type(sc) is not tuple:
            sc = (sc,)  # must be a tuple
        sc = tuple(sc_i for sc_i in sc if sc_i in sc_keys)
        a_count_axes = tuple(1 + sc_keys.index(sc_i)
                             for sc_i in sc)
        return {a: mean_grade(a_count.reshape((-1, *self.base.emf.scenario_shape)),
                              a_count_axes)
                for (a, a_count) in zip(self.base.emf.attribute_grades.keys(),
                                        self.rating_count)}

    def nap_diff(self, attr_key, sc=None, p=0.95):
        """NAP difference measure for grades of ONE attribute in
            ONE scenario dimension with EXACTLY TWO categories
        :param attr_key: single attribute key for selected result
        :param sc: scenario key, or tuple of scenario keys for result
            grade counts summed across scenario dimensions NOT included in sc
            sc[0] must be scenario dimension with exactly two categories
        :param p: (optional) scalar confidence level
        :return: dict with elements (a, nap), where
            a = attribute key, from self.emf.attribute_grades
            nap = array with
                nap[1, ...] = point estimate(s) and
                nap[[0, 2], ...] = confidence interval, stored as
            nap[:, i1, ...] = NAP difference for scenario dimension sc[0],
                given (i1, ...)-th category in scenario dimension(s) sc[1, ...]
        """
        # *** return a dict for ALL attributes, like mean_attribute_grade ??? *****
        def nap(a_count, count_axes):
            """Non-overlapping Pairs (NAP) difference measure of ONE attribute
            in ONE scenario dimension with exactly two categories
            :param a_count: mD array with
                a_count[l, k0, k1, ...] = count of l-th grade, given (k0, k1, ...)-th scenario
            :param count_axes: tuple with a_count axes to keep in result
            :return: nap = array with mean grades
                nap[i1, ...] = NAP difference, given (i1, ...)-th category
                    in self.attribute_rating count_axes, aggregated across other scenario axes
                g.shape == self.base.emf.scenario_shape
            """
            a_count = np.moveaxis(a_count, count_axes,
                                  tuple(range(1, 1 + len(count_axes))))
            sum_axis = tuple(range(1 + len(count_axes), len(a_count.shape)))
            a_count = np.sum(a_count, axis=sum_axis)  # across all other axes
            nap_u = nap_statistic(a_count[:, 1, ...], a_count[:, 0, ...], p=p)
            # *** calc CI for NAP result ? *********
            return nap_u
        # ----------------------------------------------------------------

        try:
            attr_ind = list(self.base.emf.attribute_grades.keys()).index(attr_key)
            a_count = self.rating_count[attr_ind]
        except ValueError:
            logger.warning(f'Attribute {repr(attr_key)} unknown')
            return None
        sc_keys = list(self.base.emf.scenarios.keys())
        sc_shape = self.base.emf.scenario_shape
        if sc is None:
            try:
                sc = (sc_keys[sc_shape.index(2)],)
            except ValueError:
                sc = (sc_keys[0],)
        if type(sc) is not tuple:
            sc = (sc,)  # must be a tuple, not a single key
        sc = tuple(sc_i for sc_i in sc if sc_i in sc_keys)
        if len(self.base.emf.scenarios[sc[0]]) != 2:
            logger.warning('Can calculate NAP difference only between TWO Scenario categories')
            return dict()
        a_sc_ind = tuple(1 + sc_keys.index(sc_i)
                         for sc_i in sc)
        return nap(a_count.reshape((-1, *sc_shape)),
                   a_sc_ind)


# --------------------------------------------- module help functions:
def _count_scenarios(ema_data, emf):
    """Encode EMA scenario categories as a scenario_count for analysis
    :param ema_data: list of EMA records for ONE respondent
    :param emf: ema_data.EmaFrame instance
    :return: z = 2D array with scenario_counts
        z[t, k] = number of recorded k-th <=> (k1,...)-th scenario category
        at t-th test-stage
        z.shape == (emf.n_stages, emf.n_scenarios / emf.n_stages)
    """
    z = np.zeros(emf.scenario_shape, dtype=int)
    for r in ema_data:
        z[emf.scenario_index(r)] += 1
    return z.reshape((emf.scenario_shape[0], -1))


def _count_ratings(ema_data, emf):
    """Encode EMA ATTRIBUTE attribute_grades as a rating_count list of arrays for analysis
    :param ema_data: list of EMA attribute_grades for ONE respondent
    :param emf: ema_data.EmaFrame instance
    :return: y = list of rating_count arrays,
        y[i][l, k] = number of responses at l-th ordinal level for i-th attribute,
        given the k-th <=> (k0, k1, ...)-th scenario category in the same EMA record
        y[i].shape == (len(emf.attribute_grades[i]), emf.n_scenarios)
    """
    y = [np.zeros((len(r_cat), *emf.scenario_shape), dtype=int)
         for r_cat in emf.attribute_grades.values()]
    for r in ema_data:
        sc_index = emf.scenario_index(r)
        for (i, level_i) in enumerate(emf.rating_index(r)):
            if level_i is not None:  # *** missing attribute_grades are ALLOWED
                y[i][level_i][sc_index] += 1
    for (i, r_cat) in enumerate(emf.attribute_grades.values()):
        y[i] = y[i].reshape((len(r_cat), -1))
    return y


def _initialize_scenario_param(z):
    """Crude initial estimate of scenario logprob parameters
    :param z: array with scenario counts
    :return: alpha = 2D array
        alpha[t, k] = log prob of k-th scenario at t-th test stage
        alpha.shape == (z.shape[0], z[0].size)
    """
    p = z.reshape((z.shape[0], -1)) + JEFFREYS_CONC  # PRIOR_PSEUDO_RESPONDENT
    p /= np.sum(p, axis=-1, keepdims=True)
    return np.log(p)


def _initialize_rating_theta(y):
    """Crude initial estimate of latent sensory variable for ONE attribute question
    :param y: rating_count array,
        y[l, k0, k1, ...] = number of responses at l-th ordinal level,
        given the (k0, k1, ...)-th scenario category
    :return: theta = array of sensory-variable locations
        theta[k0, k1, ...] = estimated location of latent variable,
        given the (k0, k1, ...)-th scenario category
        theta.shape == y[0].shape
    """
    n_levels = y.shape[0]
    expit_th = (np.arange(n_levels) + 0.5) / n_levels
    th = logit(expit_th)
    # back-transformed midpoints in each response interval
    p = y + JEFFREYS_CONC  # PRIOR_PSEUDO_RESPONDENT
    p /= np.sum(p, axis=0, keepdims=True)
    theta = np.tensordot(th, p, axes=1)
    # = typical (midpoint) location, given y
    return theta  # .reshape((theta.shape[0], -1))


# ------------------------------------------------- TEST:
# if __name__ == '__main__':
#     print('*** Testing _nap_statistic ***')
#     x_count = np.array([1, 2, 3, 4, 5])
#     y_count = np.array([0, 1, 2, 3, 4])
#     print(f'NAP result = {nap_statistic(x_count, y_count)}')
#



"""This package implements a Bayesian probabilistic model for
analysis of Ecological Momentary Assessment (EMA) data.

EMA methodology is used, for example,
to evaluate the subjective performance of hearing aids or other equipment,
or the effect of any other kind of psycho-social intervention,
in the everyday life of the user or client.

For example, in an EMA study of Common Sound Scenarios of hearing-aid users,
the participants might be asked to report, at random occasions in their daily life,
which type of activity they are currently engaged in,
or in which type of sound environment they are currently using their hearing aids,
and (perhaps) how important the current situation is.

The participants may also be asked to grade the performance of their current hearing aids,
for one or more perceptual Attributes, e.g., Speech Understanding and/or Sound Pleasantness.

Thus, EMA data usually include both NOMINAL and ORDINAL results.
Typically, many EMA records are collected from each participant,
but the number of records may vary a lot among respondents.

*** Main Modules:
ema_data: defines classes EmaFrame, EmaDataSet, and input functions for EMA data

ema_model: defines class EmaModel as a probabilistic model for EMA data
    and a variational learning algorithm for all model parameters

ema_group: defines class EmaGroupModel defining probabilistic model for
    ONE population and a group of EMA respondents from that population

ema_subject: defines class EmaSubjectModel defining model for
    ONE EMA respondent

ema_display: classes and functions to display analysis results

ema_base: defines help class EmaParamBase defining internal properties and methods
    for the storage indexing of all model parameters.

*** Reference:
A Leijon (2022): Bayesian Analysis of Ecological Momentary Assessment (EMA) Data.
Tech Report, unpublished, contact author Arne Leijon, e-mail: leijon@kth.se

*** Version History:
* Version 0.8.3: changed ema_group classes to avoid unnecessary data transfer to Pool processes

* Version 0.8.2: allow use multiprocessing.Pool for subjects in parallel

* Version 0.8.1: separate complete GMM for each group model,
    with mixture weights and components implemented in ema_group.EmaGroupModel

* Version 0.8: Improved variational approximation, less factorized,
    using joint q(xi_n, zeta_n) for EmaSubjectModel

* Version 0.7.2: Separate random Generators for each EmaSubjectModel

* Version 0.7.1: Minor updates:
    Allow seed for random generators, for reproducible numerical results
    Improved user control of simulated experiment, see script run_sim
    Minor change to ema_model.EmaSubjectModel.adapt_xi method

* Version 0.7:
2021-12-20, allow display of NAP effect measure of Attribute-Rating differences for
    individual participants, with point estimates and confidence intervals.
2021-12-17, display aggregated Attribute effects weighted by Scenario probabilities
2021-12-15, ema_model.EmaSubjectModel calculates individual Average Attribute Ratings, if desired.

* Version 0.6:
2021-12-08, allow user control of model restriction:
            restrict_attribute: sensory-variable location average forced -> 0.
            restrict_threshold: response-threshold median forced -> 0.

* Version 0.5.1: Minor update:
2021-11-25, allow EMA study with NO Attributes, i.e., ONLY Scenario probability profile(s)
2021-12-02, Minor fix: Attribute location == 0 given FIRST Scenario category
            regardless of regression_effect specification. (NO GOOD! Changed in v. 0.6)

* Version 0.5:
2021-10-12, Crude version, based on CountProfileCalc-2021, and PairedCompCalc (on PyPi),
2021-11-24, Functional beta version tested with simulated and (some) real data.
"""
__name__ = 'EmaCalc'
__version__ = '0.8.3'

__all__ = ['__version__', 'run_ema', 'run_sim']
